import logging
import os
import re
import sys

import sqlalchemy as sa
from alembic import command
from alembic.config import Config

from ._logger import dbos_logger


def ensure_dbos_schema(engine: sa.Engine) -> bool:
    """
    True if using DBOS migrations (DBOS schema and migrations table already exist or were created)
    False if using Alembic migrations (DBOS schema exists, but dbos_migrations table doesn't)
    """
    with engine.begin() as conn:
        # Check if dbos schema exists
        schema_result = conn.execute(
            sa.text(
                "SELECT schema_name FROM information_schema.schemata WHERE schema_name = 'dbos'"
            )
        )
        schema_existed = schema_result.fetchone() is not None

        # Create schema if it doesn't exist
        if not schema_existed:
            conn.execute(sa.text("CREATE SCHEMA dbos"))

        # Check if dbos_migrations table exists
        table_result = conn.execute(
            sa.text(
                "SELECT table_name FROM information_schema.tables WHERE table_schema = 'dbos' AND table_name = 'dbos_migrations'"
            )
        )
        table_exists = table_result.fetchone() is not None

        if table_exists:
            return True
        elif schema_existed:
            return False
        else:
            conn.execute(
                sa.text(
                    "CREATE TABLE dbos.dbos_migrations (version BIGINT NOT NULL PRIMARY KEY)"
                )
            )
            return True


def run_alembic_migrations(engine: sa.Engine) -> None:
    """Run system database schema migrations with Alembic.
    This is DEPRECATED in favor of DBOS-managed migrations.
    It is retained only for backwards compatibility and
    will be removed in the next major version."""
    # Run a schema migration for the system database
    migration_dir = os.path.join(
        os.path.dirname(os.path.realpath(__file__)), "_alembic_migrations"
    )
    alembic_cfg = Config()
    alembic_cfg.set_main_option("script_location", migration_dir)
    logging.getLogger("alembic").setLevel(logging.WARNING)
    # Alembic requires the % in URL-escaped parameters to itself be escaped to %%.
    escaped_conn_string = re.sub(
        r"%(?=[0-9A-Fa-f]{2})",
        "%%",
        engine.url.render_as_string(hide_password=False),
    )
    alembic_cfg.set_main_option("sqlalchemy.url", escaped_conn_string)
    try:
        command.upgrade(alembic_cfg, "head")
    except Exception as e:
        dbos_logger.warning(
            f"Exception during system database construction. This is most likely because the system database was configured using a later version of DBOS: {e}"
        )


def run_dbos_migrations(engine: sa.Engine) -> None:
    """Run DBOS-managed migrations by executing each SQL command in dbos_migrations."""
    with engine.begin() as conn:
        # Get current migration version
        result = conn.execute(sa.text("SELECT version FROM dbos.dbos_migrations"))
        current_version = result.fetchone()
        last_applied = current_version[0] if current_version else 0

        # Apply migrations starting from the next version
        for i, migration_sql in enumerate(dbos_migrations, 1):
            if i <= last_applied:
                continue

            # Execute the migration
            dbos_logger.info(f"Applying DBOS system database schema migration {i}")
            conn.execute(sa.text(migration_sql))

            # Update the single row with the new version
            if last_applied == 0:
                conn.execute(
                    sa.text(
                        "INSERT INTO dbos.dbos_migrations (version) VALUES (:version)"
                    ),
                    {"version": i},
                )
            else:
                conn.execute(
                    sa.text("UPDATE dbos.dbos_migrations SET version = :version"),
                    {"version": i},
                )
            last_applied = i


dbos_migration_one = """
-- Enable uuid extension for generating UUIDs
CREATE EXTENSION IF NOT EXISTS "uuid-ossp";

CREATE TABLE dbos.workflow_status (
    workflow_uuid TEXT PRIMARY KEY,
    status TEXT,
    name TEXT,
    authenticated_user TEXT,
    assumed_role TEXT,
    authenticated_roles TEXT,
    request TEXT,
    output TEXT,
    error TEXT,
    executor_id TEXT,
    created_at BIGINT NOT NULL DEFAULT (EXTRACT(epoch FROM now()) * 1000::numeric)::bigint,
    updated_at BIGINT NOT NULL DEFAULT (EXTRACT(epoch FROM now()) * 1000::numeric)::bigint,
    application_version TEXT,
    application_id TEXT,
    class_name VARCHAR(255) DEFAULT NULL,
    config_name VARCHAR(255) DEFAULT NULL,
    recovery_attempts BIGINT DEFAULT 0,
    queue_name TEXT,
    workflow_timeout_ms BIGINT,
    workflow_deadline_epoch_ms BIGINT,
    inputs TEXT,
    started_at_epoch_ms BIGINT,
    deduplication_id TEXT,
    priority INTEGER NOT NULL DEFAULT 0
);

CREATE INDEX workflow_status_created_at_index ON dbos.workflow_status (created_at);
CREATE INDEX workflow_status_executor_id_index ON dbos.workflow_status (executor_id);
CREATE INDEX workflow_status_status_index ON dbos.workflow_status (status);

ALTER TABLE dbos.workflow_status 
ADD CONSTRAINT uq_workflow_status_queue_name_dedup_id 
UNIQUE (queue_name, deduplication_id);

CREATE TABLE dbos.operation_outputs (
    workflow_uuid TEXT NOT NULL,
    function_id INTEGER NOT NULL,
    function_name TEXT NOT NULL DEFAULT '',
    output TEXT,
    error TEXT,
    child_workflow_id TEXT,
    PRIMARY KEY (workflow_uuid, function_id),
    FOREIGN KEY (workflow_uuid) REFERENCES dbos.workflow_status(workflow_uuid) 
        ON UPDATE CASCADE ON DELETE CASCADE
);

CREATE TABLE dbos.notifications (
    destination_uuid TEXT NOT NULL,
    topic TEXT,
    message TEXT NOT NULL,
    created_at_epoch_ms BIGINT NOT NULL DEFAULT (EXTRACT(epoch FROM now()) * 1000::numeric)::bigint,
    message_uuid TEXT NOT NULL DEFAULT gen_random_uuid(), -- Built-in function
    FOREIGN KEY (destination_uuid) REFERENCES dbos.workflow_status(workflow_uuid) 
        ON UPDATE CASCADE ON DELETE CASCADE
);
CREATE INDEX idx_workflow_topic ON dbos.notifications (destination_uuid, topic);

-- Create notification function
CREATE OR REPLACE FUNCTION dbos.notifications_function() RETURNS TRIGGER AS $$
DECLARE
    payload text := NEW.destination_uuid || '::' || NEW.topic;
BEGIN
    PERFORM pg_notify('dbos_notifications_channel', payload);
    RETURN NEW;
END;
$$ LANGUAGE plpgsql;

-- Create notification trigger
CREATE TRIGGER dbos_notifications_trigger
AFTER INSERT ON dbos.notifications
FOR EACH ROW EXECUTE FUNCTION dbos.notifications_function();

CREATE TABLE dbos.workflow_events (
    workflow_uuid TEXT NOT NULL,
    key TEXT NOT NULL,
    value TEXT NOT NULL,
    PRIMARY KEY (workflow_uuid, key),
    FOREIGN KEY (workflow_uuid) REFERENCES dbos.workflow_status(workflow_uuid) 
        ON UPDATE CASCADE ON DELETE CASCADE
);

-- Create events function
CREATE OR REPLACE FUNCTION dbos.workflow_events_function() RETURNS TRIGGER AS $$
DECLARE
    payload text := NEW.workflow_uuid || '::' || NEW.key;
BEGIN
    PERFORM pg_notify('dbos_workflow_events_channel', payload);
    RETURN NEW;
END;
$$ LANGUAGE plpgsql;

-- Create events trigger
CREATE TRIGGER dbos_workflow_events_trigger
AFTER INSERT ON dbos.workflow_events
FOR EACH ROW EXECUTE FUNCTION dbos.workflow_events_function();

CREATE TABLE dbos.streams (
    workflow_uuid TEXT NOT NULL,
    key TEXT NOT NULL,
    value TEXT NOT NULL,
    "offset" INTEGER NOT NULL,
    PRIMARY KEY (workflow_uuid, key, "offset"),
    FOREIGN KEY (workflow_uuid) REFERENCES dbos.workflow_status(workflow_uuid) 
        ON UPDATE CASCADE ON DELETE CASCADE
);

CREATE TABLE dbos.event_dispatch_kv (
    service_name TEXT NOT NULL,
    workflow_fn_name TEXT NOT NULL,
    key TEXT NOT NULL,
    value TEXT,
    update_seq NUMERIC(38,0),
    update_time NUMERIC(38,15),
    PRIMARY KEY (service_name, workflow_fn_name, key)
);
"""


def get_sqlite_timestamp_expr() -> str:
    """Get SQLite timestamp expression with millisecond precision for Python >= 3.12."""
    if sys.version_info >= (3, 12):
        return "(unixepoch('subsec') * 1000)"
    else:
        return "(strftime('%s','now') * 1000)"


sqlite_migration_one = f"""
CREATE TABLE workflow_status (
    workflow_uuid TEXT PRIMARY KEY,
    status TEXT,
    name TEXT,
    authenticated_user TEXT,
    assumed_role TEXT,
    authenticated_roles TEXT,
    request TEXT,
    output TEXT,
    error TEXT,
    executor_id TEXT,
    created_at INTEGER NOT NULL DEFAULT {get_sqlite_timestamp_expr()},
    updated_at INTEGER NOT NULL DEFAULT {get_sqlite_timestamp_expr()},
    application_version TEXT,
    application_id TEXT,
    class_name TEXT DEFAULT NULL,
    config_name TEXT DEFAULT NULL,
    recovery_attempts INTEGER DEFAULT 0,
    queue_name TEXT,
    workflow_timeout_ms INTEGER,
    workflow_deadline_epoch_ms INTEGER,
    inputs TEXT,
    started_at_epoch_ms INTEGER,
    deduplication_id TEXT,
    priority INTEGER NOT NULL DEFAULT 0
);

CREATE INDEX workflow_status_created_at_index ON workflow_status (created_at);
CREATE INDEX workflow_status_executor_id_index ON workflow_status (executor_id);
CREATE INDEX workflow_status_status_index ON workflow_status (status);

CREATE UNIQUE INDEX uq_workflow_status_queue_name_dedup_id 
ON workflow_status (queue_name, deduplication_id);

CREATE TABLE operation_outputs (
    workflow_uuid TEXT NOT NULL,
    function_id INTEGER NOT NULL,
    function_name TEXT NOT NULL DEFAULT '',
    output TEXT,
    error TEXT,
    child_workflow_id TEXT,
    PRIMARY KEY (workflow_uuid, function_id),
    FOREIGN KEY (workflow_uuid) REFERENCES workflow_status(workflow_uuid) 
        ON UPDATE CASCADE ON DELETE CASCADE
);

CREATE TABLE notifications (
    destination_uuid TEXT NOT NULL,
    topic TEXT,
    message TEXT NOT NULL,
    created_at_epoch_ms INTEGER NOT NULL DEFAULT {get_sqlite_timestamp_expr()},
    message_uuid TEXT NOT NULL DEFAULT (hex(randomblob(16))),
    FOREIGN KEY (destination_uuid) REFERENCES workflow_status(workflow_uuid) 
        ON UPDATE CASCADE ON DELETE CASCADE
);
CREATE INDEX idx_workflow_topic ON notifications (destination_uuid, topic);

CREATE TABLE workflow_events (
    workflow_uuid TEXT NOT NULL,
    key TEXT NOT NULL,
    value TEXT NOT NULL,
    PRIMARY KEY (workflow_uuid, key),
    FOREIGN KEY (workflow_uuid) REFERENCES workflow_status(workflow_uuid) 
        ON UPDATE CASCADE ON DELETE CASCADE
);

CREATE TABLE streams (
    workflow_uuid TEXT NOT NULL,
    key TEXT NOT NULL,
    value TEXT NOT NULL,
    "offset" INTEGER NOT NULL,
    PRIMARY KEY (workflow_uuid, key, "offset"),
    FOREIGN KEY (workflow_uuid) REFERENCES workflow_status(workflow_uuid) 
        ON UPDATE CASCADE ON DELETE CASCADE
);
"""

dbos_migrations = [dbos_migration_one]
sqlite_migrations = [sqlite_migration_one]



import threading
import traceback
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Callable

from ._logger import dbos_logger
from ._queue import Queue

if TYPE_CHECKING:
    from ._dbos import DBOSRegistry

from ._context import SetWorkflowID
from ._croniter import croniter  # type: ignore
from ._registrations import get_dbos_func_name

ScheduledWorkflow = Callable[[datetime, datetime], None]

scheduler_queue: Queue


def scheduler_loop(
    func: ScheduledWorkflow, cron: str, stop_event: threading.Event
) -> None:
    try:
        iter = croniter(cron, datetime.now(timezone.utc), second_at_beginning=True)
    except Exception as e:
        dbos_logger.error(
            f'Cannot run scheduled function {get_dbos_func_name(func)}. Invalid crontab "{cron}"'
        )
    while not stop_event.is_set():
        nextExecTime = iter.get_next(datetime)
        sleepTime = nextExecTime - datetime.now(timezone.utc)
        if stop_event.wait(timeout=sleepTime.total_seconds()):
            return
        try:
            with SetWorkflowID(
                f"sched-{get_dbos_func_name(func)}-{nextExecTime.isoformat()}"
            ):
                scheduler_queue.enqueue(func, nextExecTime, datetime.now(timezone.utc))
        except Exception:
            dbos_logger.warning(
                f"Exception encountered in scheduler thread: {traceback.format_exc()})"
            )


def scheduled(
    dbosreg: "DBOSRegistry", cron: str
) -> Callable[[ScheduledWorkflow], ScheduledWorkflow]:
    def decorator(func: ScheduledWorkflow) -> ScheduledWorkflow:
        try:
            croniter(cron, datetime.now(timezone.utc), second_at_beginning=True)
        except Exception as e:
            raise ValueError(
                f'Invalid crontab "{cron}" for scheduled function function {get_dbos_func_name(func)}.'
            )

        global scheduler_queue
        scheduler_queue = dbosreg.get_internal_queue()
        stop_event = threading.Event()
        dbosreg.register_poller(stop_event, scheduler_loop, func, cron, stop_event)
        return func

    return decorator



"""add queue dedup

Revision ID: 27ac6900c6ad
Revises: 83f3732ae8e7
Create Date: 2025-04-23 16:18:48.530047

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "27ac6900c6ad"
down_revision: Union[str, None] = "83f3732ae8e7"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    op.add_column(
        "workflow_queue",
        sa.Column(
            "deduplication_id",
            sa.Text(),
            nullable=True,
        ),
        schema="dbos",
    )

    # Unique constraint for queue_name, deduplication_id
    op.create_unique_constraint(
        "uq_workflow_queue_name_dedup_id",
        "workflow_queue",
        ["queue_name", "deduplication_id"],
        schema="dbos",
    )


def downgrade() -> None:
    op.drop_constraint(
        "uq_workflow_queue_name_dedup_id", "workflow_queue", schema="dbos"
    )
    op.drop_column("workflow_queue", "deduplication_id", schema="dbos")



import threading
import time
import uuid
from typing import Callable

import pytest
import sqlalchemy as sa

# Public API
from dbos import DBOS, Queue, SetWorkflowID
from dbos._dbos import DBOSConfiguredInstance
from dbos._error import DBOSAwaitedWorkflowCancelledError
from dbos._schemas.application_database import ApplicationSchema
from dbos._utils import INTERNAL_QUEUE_NAME, GlobalParams
from dbos._workflow_commands import garbage_collect, global_timeout
from tests.conftest import queue_entries_are_cleaned_up


def test_cancel_resume(dbos: DBOS) -> None:
    steps_completed = 0
    workflow_event = threading.Event()
    main_thread_event = threading.Event()
    input = 5

    @DBOS.step()
    def step_one() -> None:
        nonlocal steps_completed
        steps_completed += 1

    @DBOS.step()
    def step_two() -> None:
        nonlocal steps_completed
        steps_completed += 1

    @DBOS.workflow()
    def simple_workflow(x: int) -> int:
        step_one()
        main_thread_event.set()
        workflow_event.wait()
        # A handler like this should not catch DBOSWorkflowCancelledError
        try:
            step_two()
        except Exception:
            raise
        return x

    # Start the workflow and cancel it.
    # Verify it stops after step one but before step two
    wfid = str(uuid.uuid4())
    with SetWorkflowID(wfid):
        handle = DBOS.start_workflow(simple_workflow, input)
    main_thread_event.wait()
    DBOS.cancel_workflow(wfid)
    workflow_event.set()
    with pytest.raises(DBOSAwaitedWorkflowCancelledError):
        handle.get_result()
    assert steps_completed == 1

    # Resume the workflow. Verify it completes successfully.
    handle = DBOS.resume_workflow(wfid)
    assert handle.get_status().app_version == GlobalParams.app_version
    assert handle.get_status().queue_name == INTERNAL_QUEUE_NAME
    assert handle.get_result() == input
    assert steps_completed == 2

    # Resume the workflow again. Verify it does not run again.
    handle = DBOS.resume_workflow(wfid)
    assert handle.get_result() == input
    assert steps_completed == 2

    assert queue_entries_are_cleaned_up(dbos)


def test_cancel_resume_txn(dbos: DBOS) -> None:
    txn_completed = 0
    workflow_event = threading.Event()
    main_thread_event = threading.Event()
    input = 5

    @DBOS.transaction()
    def txn_one() -> None:
        nonlocal txn_completed
        txn_completed += 1

    @DBOS.transaction()
    def txn_two() -> None:
        nonlocal txn_completed
        txn_completed += 1

    @DBOS.workflow()
    def simple_workflow(x: int) -> int:
        txn_one()
        main_thread_event.set()
        workflow_event.wait()
        txn_two()
        return x

    # Start the workflow and cancel it.
    # Verify it stops after step one but before step two
    wfid = str(uuid.uuid4())
    with SetWorkflowID(wfid):
        handle = DBOS.start_workflow(simple_workflow, input)
    main_thread_event.wait()
    DBOS.cancel_workflow(wfid)
    workflow_event.set()
    with pytest.raises(DBOSAwaitedWorkflowCancelledError):
        handle.get_result()
    assert txn_completed == 1

    # Resume the workflow. Verify it completes successfully.
    handle = DBOS.resume_workflow(wfid)
    assert handle.get_result() == input
    assert txn_completed == 2

    # Resume the workflow again. Verify it does not run again.
    handle = DBOS.resume_workflow(wfid)
    assert handle.get_result() == input
    assert txn_completed == 2


def test_cancel_resume_queue(dbos: DBOS) -> None:
    steps_completed = 0
    workflow_event = threading.Event()
    main_thread_event = threading.Event()
    input = 5

    queue = Queue("test_queue")

    @DBOS.step()
    def step_one() -> None:
        nonlocal steps_completed
        steps_completed += 1

    @DBOS.step()
    def step_two() -> None:
        nonlocal steps_completed
        steps_completed += 1

    @DBOS.workflow()
    def simple_workflow(x: int) -> int:
        step_one()
        main_thread_event.set()
        workflow_event.wait()
        step_two()
        return x

    # Start the workflow and cancel it.
    # Verify it stops after step one but before step two
    wfid = str(uuid.uuid4())
    with SetWorkflowID(wfid):
        handle = queue.enqueue(simple_workflow, input)
    main_thread_event.wait()
    DBOS.cancel_workflow(wfid)
    workflow_event.set()
    with pytest.raises(Exception):
        handle.get_result()
    assert steps_completed == 1

    # Resume the workflow. Verify it completes successfully.
    handle = DBOS.resume_workflow(wfid)
    assert handle.get_result() == input
    assert steps_completed == 2

    # Resume the workflow again. Verify it does not run again.
    handle = DBOS.resume_workflow(wfid)
    assert handle.get_result() == input
    assert steps_completed == 2

    # Verify nothing is left on any queue
    assert queue_entries_are_cleaned_up(dbos)


def test_restart(dbos: DBOS) -> None:
    input = 2
    multiplier = 5

    @DBOS.dbos_class()
    class TestClass(DBOSConfiguredInstance):

        def __init__(self, multiplier: int) -> None:
            self.multiply: Callable[[int], int] = lambda x: x * multiplier
            super().__init__("test_class")

        @DBOS.workflow()
        def workflow(self, x: int) -> int:
            return self.multiply(x)

        @DBOS.step()
        def step(self, x: int) -> int:
            return self.multiply(x)

    inst = TestClass(multiplier)

    # Start the workflow, let it finish, restart it.
    # Verify it returns the same result with a different workflow ID.
    handle = DBOS.start_workflow(inst.workflow, input)
    assert handle.get_result() == input * multiplier
    forked_handle = DBOS.restart_workflow(handle.workflow_id)
    assert forked_handle.workflow_id != handle.workflow_id
    assert forked_handle.get_result() == input * multiplier

    # Enqueue the workflow, let it finish, restart it.
    # Verify it returns the same result with a different workflow ID and queue.
    queue = Queue("test_queue")
    handle = queue.enqueue(inst.workflow, input)
    assert handle.get_result() == input * multiplier
    forked_handle = DBOS.restart_workflow(handle.workflow_id)
    assert forked_handle.workflow_id != handle.workflow_id
    assert forked_handle.get_status().queue_name == INTERNAL_QUEUE_NAME
    assert forked_handle.get_result() == input * multiplier

    # Enqueue the step, let it finish, restart it.
    # Verify it returns the same result with a different workflow ID and queue.
    handle = queue.enqueue(inst.step, input)
    assert handle.get_result() == input * multiplier
    forked_handle = DBOS.restart_workflow(handle.workflow_id)
    assert forked_handle.workflow_id != handle.workflow_id
    assert forked_handle.get_status().queue_name != handle.get_status().queue_name
    assert forked_handle.get_result() == input * multiplier

    # Verify restarting a nonexistent workflow throws an exception
    with pytest.raises(Exception):
        DBOS.restart_workflow("fake_id")

    # Verify nothing is left on any queue
    assert queue_entries_are_cleaned_up(dbos)


def test_fork_steps(
    dbos: DBOS,
) -> None:

    stepOneCount = 0
    stepTwoCount = 0
    stepThreeCount = 0
    stepFourCount = 0
    stepFiveCount = 0

    @DBOS.workflow()
    def simple_workflow(x: int) -> int:
        return stepOne(x) + stepTwo(x) + stepThree(x) + stepFour(x) + stepFive(x)

    @DBOS.step()
    def stepOne(x: int) -> int:
        nonlocal stepOneCount
        stepOneCount += 1
        return x + 1

    @DBOS.step()
    def stepTwo(x: int) -> int:
        nonlocal stepTwoCount
        stepTwoCount += 1
        return x + 2

    @DBOS.step()
    def stepThree(x: int) -> int:
        nonlocal stepThreeCount
        stepThreeCount += 1
        return x + 3

    @DBOS.step()
    def stepFour(x: int) -> int:
        nonlocal stepFourCount
        stepFourCount += 1
        return x + 4

    @DBOS.step()
    def stepFive(x: int) -> int:
        nonlocal stepFiveCount
        stepFiveCount += 1
        return x + 5

    input = 1
    output = 5 * input + 15

    wfid = str(uuid.uuid4())
    with SetWorkflowID(wfid):
        assert simple_workflow(input) == output

    assert stepOneCount == 1
    assert stepTwoCount == 1
    assert stepThreeCount == 1
    assert stepFourCount == 1
    assert stepFiveCount == 1

    fork_id = str(uuid.uuid4())
    with SetWorkflowID(fork_id):
        forked_handle = DBOS.fork_workflow(wfid, 3)
    assert forked_handle.workflow_id == fork_id
    assert forked_handle.get_status().app_version == GlobalParams.app_version
    assert forked_handle.get_result() == output

    assert stepOneCount == 1
    assert stepTwoCount == 1
    assert stepThreeCount == 2
    assert stepFourCount == 2
    assert stepFiveCount == 2

    forked_handle = DBOS.fork_workflow(wfid, 5)
    assert forked_handle.workflow_id != wfid
    assert forked_handle.get_result() == output

    assert stepOneCount == 1
    assert stepTwoCount == 1
    assert stepThreeCount == 2
    assert stepFourCount == 2
    assert stepFiveCount == 3

    forked_handle = DBOS.fork_workflow(wfid, 1)
    assert forked_handle.workflow_id != wfid
    assert forked_handle.get_result() == output

    assert stepOneCount == 2
    assert stepTwoCount == 2
    assert stepThreeCount == 3
    assert stepFourCount == 3
    assert stepFiveCount == 4


def test_restart_fromsteps_transactionsonly(
    dbos: DBOS,
) -> None:

    trOneCount = 0
    trTwoCount = 0
    trThreeCount = 0
    trFourCount = 0
    trFiveCount = 0

    @DBOS.workflow()
    def simple_workflow() -> None:
        trOne()
        trTwo()
        trThree()
        trFour()
        trFive()
        return

    @DBOS.transaction()
    def trOne() -> None:
        nonlocal trOneCount
        trOneCount += 1
        return

    @DBOS.transaction()
    def trTwo() -> None:
        nonlocal trTwoCount
        trTwoCount += 1
        return

    @DBOS.transaction()
    def trThree() -> None:
        nonlocal trThreeCount
        trThreeCount += 1
        return

    @DBOS.transaction()
    def trFour() -> None:
        nonlocal trFourCount
        trFourCount += 1
        return

    @DBOS.transaction()
    def trFive() -> None:
        nonlocal trFiveCount
        trFiveCount += 1
        return

    wfid = str(uuid.uuid4())
    with SetWorkflowID(wfid):
        simple_workflow()

    assert trOneCount == 1
    assert trTwoCount == 1
    assert trThreeCount == 1
    assert trFourCount == 1
    assert trFiveCount == 1

    forked_handle = DBOS.fork_workflow(wfid, 2)
    assert forked_handle.workflow_id != wfid
    forked_handle.get_result()

    assert trOneCount == 1
    assert trTwoCount == 2
    assert trThreeCount == 2
    assert trFourCount == 2
    assert trFiveCount == 2

    forked_handle = DBOS.fork_workflow(wfid, 4)
    assert forked_handle.workflow_id != wfid
    forked_handle.get_result()

    assert trOneCount == 1
    assert trTwoCount == 2
    assert trThreeCount == 2
    assert trFourCount == 3
    assert trFiveCount == 3

    forked_handle = DBOS.fork_workflow(wfid, 1)
    assert forked_handle.workflow_id != wfid
    forked_handle.get_result()

    assert trOneCount == 2
    assert trTwoCount == 3
    assert trThreeCount == 3
    assert trFourCount == 4
    assert trFiveCount == 4


def test_restart_fromsteps_steps_tr(
    dbos: DBOS,
) -> None:

    trOneCount = 0
    stepTwoCount = 0
    trThreeCount = 0
    stepFourCount = 0
    trFiveCount = 0

    @DBOS.workflow()
    def simple_workflow() -> None:
        trOne()
        stepTwo()
        trThree()
        stepFour()
        trFive()
        return

    @DBOS.transaction()
    def trOne() -> None:
        nonlocal trOneCount
        trOneCount += 1
        return

    @DBOS.step()
    def stepTwo() -> None:
        nonlocal stepTwoCount
        stepTwoCount += 1
        return

    @DBOS.transaction()
    def trThree() -> None:
        nonlocal trThreeCount
        trThreeCount += 1
        return

    @DBOS.step()
    def stepFour() -> None:
        nonlocal stepFourCount
        stepFourCount += 1
        return

    @DBOS.transaction()
    def trFive() -> None:
        nonlocal trFiveCount
        trFiveCount += 1
        return

    wfid = str(uuid.uuid4())
    with SetWorkflowID(wfid):
        simple_workflow()

    assert trOneCount == 1
    assert stepTwoCount == 1
    assert trThreeCount == 1
    assert stepFourCount == 1
    assert trFiveCount == 1

    forked_handle = DBOS.fork_workflow(wfid, 3)
    assert forked_handle.workflow_id != wfid
    forked_handle.get_result()

    assert trOneCount == 1
    assert stepTwoCount == 1
    assert trThreeCount == 2
    assert stepFourCount == 2
    assert trFiveCount == 2

    forked_handle = DBOS.fork_workflow(wfid, 5)
    assert forked_handle.workflow_id != wfid
    forked_handle.get_result()

    assert trOneCount == 1
    assert stepTwoCount == 1
    assert trThreeCount == 2
    assert stepFourCount == 2
    assert trFiveCount == 3

    # invalid < 1 will default to 1
    forked_handle = DBOS.fork_workflow(wfid, -1)
    assert forked_handle.workflow_id != wfid
    forked_handle.get_result()

    assert trOneCount == 2
    assert stepTwoCount == 2
    assert trThreeCount == 3
    assert stepFourCount == 3
    assert trFiveCount == 4


def test_restart_fromsteps_childwf(
    dbos: DBOS,
) -> None:

    stepOneCount = 0
    childwfCount = 0
    stepThreeCount = 0

    @DBOS.workflow()
    def simple_workflow() -> None:
        stepOne()
        wfid = str(uuid.uuid4())
        with SetWorkflowID(wfid):
            handle = dbos.start_workflow(
                child_workflow,
                wfid,
            )
        handle.get_result()
        stepThree()
        return

    @DBOS.step()
    def stepOne() -> None:
        nonlocal stepOneCount
        stepOneCount += 1
        return

    @DBOS.workflow()
    def child_workflow(id: str) -> str:
        nonlocal childwfCount
        childwfCount += 1
        return id

    @DBOS.step()
    def stepThree() -> None:
        nonlocal stepThreeCount
        stepThreeCount += 1
        return

    @DBOS.workflow()
    def fork(workflow_id: str, step: int) -> str:
        handle = DBOS.fork_workflow(workflow_id, step)
        handle.get_result()
        return handle.get_workflow_id()

    wfid = str(uuid.uuid4())
    with SetWorkflowID(wfid):
        h = DBOS.start_workflow(simple_workflow)
    h.get_result()

    assert stepOneCount == 1
    assert childwfCount == 1
    assert stepThreeCount == 1

    forked_handle = DBOS.fork_workflow(wfid, 2)
    forked_handle.get_result()
    assert forked_handle.workflow_id != wfid
    assert stepOneCount == 1
    assert childwfCount == 2
    assert stepThreeCount == 2

    forked_handle = DBOS.fork_workflow(wfid, 3)
    forked_handle.get_result()
    assert forked_handle.workflow_id != wfid
    assert stepOneCount == 1
    assert childwfCount == 2
    assert stepThreeCount == 3

    # call fork from within a workflow
    forkwfid = str(uuid.uuid4())
    with SetWorkflowID(forkwfid):
        fh = DBOS.start_workflow(fork, wfid, 1)
    firstforkedwfid = fh.get_result()
    assert firstforkedwfid != wfid
    assert stepOneCount == 2
    assert childwfCount == 3
    assert stepThreeCount == 4

    # call the workflow again with the same id
    # testing that fork is not called again
    with SetWorkflowID(forkwfid):
        fh2 = DBOS.start_workflow(fork, wfid, 1)

    secondforkedwfid = fh2.get_result()
    assert secondforkedwfid == firstforkedwfid

    assert stepOneCount == 2
    assert childwfCount == 3
    assert stepThreeCount == 4


def test_fork_version(
    dbos: DBOS,
) -> None:

    stepOneCount = 0
    stepTwoCount = 0

    @DBOS.workflow()
    def simple_workflow(x: int) -> int:
        return stepOne(x) + stepTwo(x)

    @DBOS.step()
    def stepOne(x: int) -> int:
        nonlocal stepOneCount
        stepOneCount += 1
        return x + 1

    @DBOS.step()
    def stepTwo(x: int) -> int:
        nonlocal stepTwoCount
        stepTwoCount += 1
        return x + 2

    input = 1
    output = 2 * input + 3

    workflow_id = str(uuid.uuid4())
    with SetWorkflowID(workflow_id):
        assert simple_workflow(input) == output

    # Fork the workflow with a different version. Verify it is set to that version.
    new_version = "my_new_version"
    handle = DBOS.fork_workflow(workflow_id, 2, application_version=new_version)
    assert handle.get_status().app_version == new_version
    assert handle.get_status().queue_name == INTERNAL_QUEUE_NAME
    # Set the global version to this new version, verify the workflow completes
    GlobalParams.app_version = new_version
    assert handle.get_result() == output
    assert queue_entries_are_cleaned_up(dbos)


def test_garbage_collection(dbos: DBOS, skip_with_sqlite_imprecise_time: None) -> None:
    event = threading.Event()

    @DBOS.step()
    def step(x: int) -> int:
        return x

    @DBOS.transaction()
    def txn(x: int) -> int:
        DBOS.sql_session.execute(sa.text("SELECT 1")).fetchall()
        return x

    @DBOS.workflow()
    def workflow(x: int) -> int:
        step(x)
        txn(x)
        return x

    @DBOS.workflow()
    def blocked_workflow() -> str:
        txn(0)
        event.wait()
        workflow_id = DBOS.workflow_id
        assert workflow_id is not None
        return workflow_id

    num_workflows = 10

    handle = DBOS.start_workflow(blocked_workflow)
    for i in range(num_workflows):
        assert workflow(i) == i

    # Garbage collect all but one workflow
    garbage_collect(dbos, cutoff_epoch_timestamp_ms=None, rows_threshold=1)
    # Verify two workflows remain: the newest and the blocked workflow
    workflows = DBOS.list_workflows()
    assert len(workflows) == 2
    assert workflows[0].workflow_id == handle.workflow_id
    # Verify txn outputs are preserved only for the remaining workflows
    with dbos._app_db.engine.begin() as c:
        rows = c.execute(
            sa.select(
                ApplicationSchema.transaction_outputs.c.workflow_uuid,
            )
        ).all()
        assert len(rows) == 2

    # Garbage collect all previous workflows
    garbage_collect(
        dbos, cutoff_epoch_timestamp_ms=int(time.time() * 1000), rows_threshold=None
    )
    # Verify only the blocked workflow remains
    workflows = DBOS.list_workflows()
    assert len(workflows) == 1
    assert workflows[0].workflow_id == handle.workflow_id
    # Verify txn outputs are preserved only for the remaining workflow
    with dbos._app_db.engine.begin() as c:
        rows = c.execute(
            sa.select(
                ApplicationSchema.transaction_outputs.c.workflow_uuid,
            )
        ).all()
        assert len(rows) == 1

    # Finish the blocked workflow, garbage collect everything
    event.set()
    assert handle.get_result() is not None
    garbage_collect(
        dbos, cutoff_epoch_timestamp_ms=int(time.time() * 1000), rows_threshold=None
    )
    # Verify only the blocked workflow remains
    workflows = DBOS.list_workflows()
    assert len(workflows) == 0

    # Verify GC runs without error on a blank table
    garbage_collect(dbos, cutoff_epoch_timestamp_ms=None, rows_threshold=1)

    # Run workflows, wait, run them again
    for i in range(num_workflows):
        assert workflow(i) == i
    time.sleep(1)
    for i in range(num_workflows):
        assert workflow(i) == i

    # GC the first half, verify only half were GC'ed
    garbage_collect(
        dbos,
        cutoff_epoch_timestamp_ms=int(time.time() * 1000) - 1000,
        rows_threshold=None,
    )
    workflows = DBOS.list_workflows()
    assert len(workflows) == num_workflows


def test_global_timeout(dbos: DBOS) -> None:
    event = threading.Event()

    @DBOS.workflow()
    def blocked_workflow() -> str:
        while not event.wait(0):
            DBOS.sleep(0.1)
        workflow_id = DBOS.workflow_id
        assert workflow_id is not None
        return workflow_id

    num_workflows = 10
    handles = [DBOS.start_workflow(blocked_workflow) for _ in range(num_workflows)]

    # Wait two seconds, start one final workflow, then timeout all workflows started more than one second ago
    time.sleep(2)
    final_handle = DBOS.start_workflow(blocked_workflow)
    cutoff_epoch_timestamp_ms = int(time.time() * 1000) - 1000
    global_timeout(dbos, cutoff_epoch_timestamp_ms)

    # Verify all workflows started before the global timeout are cancelled
    for handle in handles:
        with pytest.raises(DBOSAwaitedWorkflowCancelledError):
            handle.get_result()
    event.set()
    assert final_handle.get_result() is not None



import os
import platform
import signal
import subprocess
import time
import typing
from os import path
from typing import Any, Optional, Tuple

import jsonpickle  # type: ignore
import sqlalchemy as sa
import typer
from rich import print as richprint
from rich.prompt import IntPrompt
from typing_extensions import Annotated, List

from dbos._context import SetWorkflowID
from dbos._debug import debug_workflow, parse_start_command
from dbos.cli.migration import grant_dbos_schema_permissions, migrate_dbos_databases

from .._client import DBOSClient
from .._dbos_config import (
    ConfigFile,
    _app_name_to_db_name,
    _is_valid_app_name,
    get_application_database_url,
    get_system_database_url,
    load_config,
)
from .._docker_pg_helper import start_docker_pg, stop_docker_pg
from .._sys_db import SystemDatabase
from .._utils import GlobalParams
from ..cli._github_init import create_template_from_github
from ._template_init import copy_template, get_project_name, get_templates_directory


def _get_db_url(
    *, system_database_url: Optional[str], application_database_url: Optional[str]
) -> Tuple[str, str]:
    """
    Get the database URL to use for the DBOS application.
    Order of precedence:
    - In DBOS Cloud, use the environment variables provided.
    - Use database URL arguments if provided.
    - If the `dbos-config.yaml` file is present, use the database URLs from it.

    Otherwise fallback to the same SQLite Postgres URL than the DBOS library.
    Note that for the latter to be possible, a configuration file must have been found, with an application name set.
    """
    if os.environ.get("DBOS__CLOUD") == "true":
        system_database_url = os.environ.get("DBOS_SYSTEM_DATABASE_URL")
        application_database_url = os.environ.get("DBOS_DATABASE_URL")
        assert system_database_url and application_database_url
        return system_database_url, application_database_url
    if system_database_url or application_database_url:
        cfg: ConfigFile = {
            "system_database_url": system_database_url,
            "database_url": application_database_url,
        }
        return get_system_database_url(cfg), get_application_database_url(cfg)
    else:
        # Load from config file if present
        try:
            config = load_config(silent=True)
            if config.get("database_url") or config.get("system_database_url"):
                return get_system_database_url(config), get_application_database_url(
                    config
                )
            else:
                _app_db_name = _app_name_to_db_name(config["name"])
                # Fallback on the same defaults than the DBOS library
                default_url = f"sqlite:///{_app_db_name}.sqlite"
                return default_url, default_url
        except (FileNotFoundError, OSError):
            typer.echo(
                f"Error: Missing database URL: please set it using CLI flags or your dbos-config.yaml file.",
                err=True,
            )
            raise typer.Exit(code=1)


app = typer.Typer()


@app.command(help="Show the version and exit")
def version() -> None:
    """Display the current version of DBOS CLI."""
    typer.echo(f"DBOS CLI version: {GlobalParams.dbos_version}")


workflow = typer.Typer()
queue = typer.Typer()

app.add_typer(workflow, name="workflow", help="Manage DBOS workflows")
workflow.add_typer(queue, name="queue", help="Manage enqueued workflows")

postgres = typer.Typer()
app.add_typer(
    postgres, name="postgres", help="Manage local Postgres database with Docker"
)


@postgres.command(name="start", help="Start a local Postgres database")
def pg_start() -> None:
    start_docker_pg()


@postgres.command(name="stop", help="Stop the local Postgres database")
def pg_stop() -> None:
    stop_docker_pg()


def _on_windows() -> bool:
    return platform.system() == "Windows"


@app.command(
    help="Start your DBOS application using the start commands in 'dbos-config.yaml'"
)
def start() -> None:
    config = load_config(silent=True)
    start_commands = config["runtimeConfig"]["start"]
    typer.echo("Executing start commands from 'dbos-config.yaml'")
    for command in start_commands:
        typer.echo(f"Executing: {command}")

        # Run the command in the child process.
        # On Unix-like systems, set its process group
        process = subprocess.Popen(
            command,
            shell=True,
            text=True,
            preexec_fn=os.setsid if not _on_windows() else None,
        )

        def signal_handler(signum: int, frame: Any) -> None:
            """
            Forward kill signals to children.

            When we receive a signal, send it to the entire process group of the child.
            If that doesn't work, SIGKILL them then exit.
            """
            # Send the signal to the child's entire process group
            if process.poll() is None:
                os.killpg(os.getpgid(process.pid), signum)

            # Give some time for the child to terminate
            for _ in range(10):  # Wait up to 1 second
                if process.poll() is not None:
                    break
                time.sleep(0.1)

            # If the child is still running, force kill it
            if process.poll() is None:
                try:
                    os.killpg(os.getpgid(process.pid), signal.SIGKILL)
                except Exception:
                    pass

            # Exit immediately
            os._exit(process.returncode if process.returncode is not None else 1)

        # Configure the single handler only on Unix-like systems.
        # TODO: Also kill the children on Windows.
        if not _on_windows():
            signal.signal(signal.SIGINT, signal_handler)
            signal.signal(signal.SIGTERM, signal_handler)
        process.wait()


@app.command(help="Initialize a new DBOS application from a template")
def init(
    project_name: Annotated[
        typing.Optional[str], typer.Argument(help="Specify application name")
    ] = None,
    template: Annotated[
        typing.Optional[str],
        typer.Option("--template", "-t", help="Specify template to use"),
    ] = None,
    config: Annotated[
        bool,
        typer.Option("--config", "-c", help="Only add dbos-config.yaml"),
    ] = False,
) -> None:
    try:
        git_templates = ["dbos-toolbox", "dbos-app-starter", "dbos-cron-starter"]
        templates_dir = get_templates_directory()

        project_name, template = _resolve_project_name_and_template(
            project_name=project_name,
            template=template,
            config=config,
            git_templates=git_templates,
            templates_dir=templates_dir,
        )

        if template in git_templates:
            create_template_from_github(app_name=project_name, template_name=template)
        else:
            copy_template(
                path.join(templates_dir, template), project_name, config_mode=config
            )
    except Exception as e:
        richprint(f"[red]{e}[/red]")


def _resolve_project_name_and_template(
    project_name: Optional[str],
    template: Optional[str],
    config: bool,
    git_templates: List[str],
    templates_dir: str,
) -> tuple[str, str]:
    templates = git_templates + [
        x.name for x in os.scandir(templates_dir) if x.is_dir()
    ]

    if config and template is None:
        template = templates[-1]

    if template:
        if template not in templates:
            raise Exception(f"Template {template} not found in {templates_dir}")
    else:
        richprint("\n[bold]Available templates:[/bold]")
        for idx, template_name in enumerate(templates, 1):
            richprint(f"  {idx}. {template_name}")
        while True:
            try:
                choice = IntPrompt.ask(
                    "\nSelect template number",
                    show_choices=False,
                    show_default=False,
                )
                if 1 <= choice <= len(templates):
                    template = templates[choice - 1]
                    break
                else:
                    richprint(
                        "[red]Invalid selection. Please choose a number from the list.[/red]"
                    )
            except (KeyboardInterrupt, EOFError):
                raise typer.Abort()
            except ValueError:
                richprint("[red]Please enter a valid number.[/red]")

    if template in git_templates:
        if project_name is None:
            project_name = template
    else:
        if project_name is None:
            project_name = typing.cast(
                str,
                typer.prompt("What is your project's name?", get_project_name()),
            )

    if not _is_valid_app_name(project_name):
        raise Exception(
            f"{project_name} is an invalid DBOS app name. App names must be between 3 and 30 characters long and contain only lowercase letters, numbers, dashes, and underscores."
        )

    assert project_name is not None, "Project name cannot be None"
    assert template is not None, "Template name cannot be None"

    return project_name, template


@app.command(help="Create DBOS system tables.")
def migrate(
    application_database_url: Annotated[
        typing.Optional[str],
        typer.Option(
            "--db-url",
            "-D",
            help="Your DBOS application database URL",
        ),
    ] = None,
    system_database_url: Annotated[
        typing.Optional[str],
        typer.Option(
            "--sys-db-url",
            "-s",
            help="Your DBOS system database URL",
        ),
    ] = None,
    application_role: Annotated[
        typing.Optional[str],
        typer.Option(
            "--app-role",
            "-r",
            help="The role with which you will run your DBOS application",
        ),
    ] = None,
) -> None:
    system_database_url, application_database_url = _get_db_url(
        system_database_url=system_database_url,
        application_database_url=application_database_url,
    )

    typer.echo(f"Starting DBOS migrations")
    typer.echo(f"Application database: {sa.make_url(application_database_url)}")
    typer.echo(f"System database: {sa.make_url(system_database_url)}")

    # First, run DBOS migrations on the system database and the application database
    migrate_dbos_databases(
        app_database_url=application_database_url,
        system_database_url=system_database_url,
    )

    # Next, assign permissions on the DBOS schema to the application role, if any
    if application_role:
        grant_dbos_schema_permissions(
            database_url=application_database_url, role_name=application_role
        )
        grant_dbos_schema_permissions(
            database_url=system_database_url, role_name=application_role
        )

    # Next, run any custom migration commands specified in the configuration
    if os.path.exists("dbos-config.yaml"):
        config = load_config(silent=True)
        if "database" not in config:
            config["database"] = {}
        migrate_commands = (
            config["database"]["migrate"]
            if "migrate" in config["database"] and config["database"]["migrate"]
            else []
        )
        if migrate_commands:
            typer.echo("Executing migration commands from 'dbos-config.yaml'")
            try:
                for command in migrate_commands:
                    typer.echo(f"Executing migration command: {command}")
                    result = subprocess.run(command, shell=True, text=True)
                    if result.returncode != 0:
                        typer.echo(f"Migration command failed: {command}")
                        typer.echo(result.stderr)
                        raise typer.Exit(1)
                    if result.stdout:
                        typer.echo(result.stdout.rstrip())
            except Exception as e:
                typer.echo(f"An error occurred during schema migration: {e}")
                raise typer.Exit(code=1)


@app.command(help="Reset the DBOS system database")
def reset(
    yes: bool = typer.Option(False, "-y", "--yes", help="Skip confirmation prompt"),
    application_database_url: Annotated[
        typing.Optional[str],
        typer.Option(
            "--db-url",
            "-D",
            help="Your DBOS application database URL",
        ),
    ] = None,
    system_database_url: Annotated[
        typing.Optional[str],
        typer.Option(
            "--sys-db-url",
            "-s",
            help="Your DBOS system database URL",
        ),
    ] = None,
) -> None:
    if not yes:
        confirm = typer.confirm(
            "This command resets your DBOS system database, deleting metadata about past workflows and steps. Are you sure you want to proceed?"
        )
        if not confirm:
            typer.echo("Operation cancelled.")
            raise typer.Exit()
    try:
        system_database_url, application_database_url = _get_db_url(
            system_database_url=system_database_url,
            application_database_url=application_database_url,
        )
        SystemDatabase.reset_system_database(system_database_url)
    except Exception as e:
        typer.echo(f"Error resetting system database: {str(e)}")
        return


@app.command(help="Replay Debug a DBOS workflow")
def debug(
    workflow_id: Annotated[str, typer.Argument(help="Workflow ID to debug")],
) -> None:
    config = load_config(silent=True)
    start = config["runtimeConfig"]["start"]
    if not start:
        typer.echo("No start commands found in 'dbos-config.yaml'")
        raise typer.Exit(code=1)
    if len(start) > 1:
        typer.echo("Multiple start commands found in 'dbos-config.yaml'")
        raise typer.Exit(code=1)
    entrypoint = parse_start_command(start[0])
    debug_workflow(workflow_id, entrypoint)


@workflow.command(help="List workflows for your application")
def list(
    application_database_url: Annotated[
        typing.Optional[str],
        typer.Option(
            "--db-url",
            "-D",
            help="Your DBOS application database URL",
        ),
    ] = None,
    system_database_url: Annotated[
        typing.Optional[str],
        typer.Option(
            "--sys-db-url",
            "-s",
            help="Your DBOS system database URL",
        ),
    ] = None,
    limit: Annotated[
        int,
        typer.Option("--limit", "-l", help="Limit the results returned"),
    ] = 10,
    user: Annotated[
        typing.Optional[str],
        typer.Option("--user", "-u", help="Retrieve workflows run by this user"),
    ] = None,
    starttime: Annotated[
        typing.Optional[str],
        typer.Option(
            "--start-time",
            "-t",
            help="Retrieve workflows starting after this timestamp (ISO 8601 format)",
        ),
    ] = None,
    endtime: Annotated[
        typing.Optional[str],
        typer.Option(
            "--end-time",
            "-e",
            help="Retrieve workflows starting before this timestamp (ISO 8601 format)",
        ),
    ] = None,
    status: Annotated[
        typing.Optional[str],
        typer.Option(
            "--status",
            "-S",
            help="Retrieve workflows with this status (PENDING, SUCCESS, ERROR, ENQUEUED, CANCELLED, or MAX_RECOVERY_ATTEMPTS_EXCEEDED)",
        ),
    ] = None,
    appversion: Annotated[
        typing.Optional[str],
        typer.Option(
            "--application-version",
            "-v",
            help="Retrieve workflows with this application version",
        ),
    ] = None,
    name: Annotated[
        typing.Optional[str],
        typer.Option(
            "--name",
            "-n",
            help="Retrieve workflows with this name",
        ),
    ] = None,
    sort_desc: Annotated[
        bool,
        typer.Option(
            "--sort-desc",
            "-d",
            help="Sort the results in descending order (older first)",
        ),
    ] = False,
    offset: Annotated[
        typing.Optional[int],
        typer.Option(
            "--offset",
            "-o",
            help="Offset for pagination",
        ),
    ] = None,
) -> None:
    system_database_url, application_database_url = _get_db_url(
        system_database_url=system_database_url,
        application_database_url=application_database_url,
    )
    client = DBOSClient(
        application_database_url=application_database_url,
        system_database_url=system_database_url,
    )
    workflows = client.list_workflows(
        limit=limit,
        offset=offset,
        sort_desc=sort_desc,
        user=user,
        start_time=starttime,
        end_time=endtime,
        status=status,
        app_version=appversion,
        name=name,
    )
    print(jsonpickle.encode(workflows, unpicklable=False))


@workflow.command(help="Retrieve the status of a workflow")
def get(
    workflow_id: Annotated[str, typer.Argument()],
    application_database_url: Annotated[
        typing.Optional[str],
        typer.Option(
            "--db-url",
            "-D",
            help="Your DBOS application database URL",
        ),
    ] = None,
    system_database_url: Annotated[
        typing.Optional[str],
        typer.Option(
            "--sys-db-url",
            "-s",
            help="Your DBOS system database URL",
        ),
    ] = None,
) -> None:
    system_database_url, application_database_url = _get_db_url(
        system_database_url=system_database_url,
        application_database_url=application_database_url,
    )
    client = DBOSClient(
        application_database_url=application_database_url,
        system_database_url=system_database_url,
    )
    status = client.retrieve_workflow(workflow_id=workflow_id).get_status()
    print(jsonpickle.encode(status, unpicklable=False))


@workflow.command(help="List the steps of a workflow")
def steps(
    workflow_id: Annotated[str, typer.Argument()],
    application_database_url: Annotated[
        typing.Optional[str],
        typer.Option(
            "--db-url",
            "-D",
            help="Your DBOS application database URL",
        ),
    ] = None,
    system_database_url: Annotated[
        typing.Optional[str],
        typer.Option(
            "--sys-db-url",
            "-s",
            help="Your DBOS system database URL",
        ),
    ] = None,
) -> None:
    system_database_url, application_database_url = _get_db_url(
        system_database_url=system_database_url,
        application_database_url=application_database_url,
    )
    client = DBOSClient(
        application_database_url=application_database_url,
        system_database_url=system_database_url,
    )
    print(
        jsonpickle.encode(
            client.list_workflow_steps(workflow_id=workflow_id),
            unpicklable=False,
        )
    )


@workflow.command(
    help="Cancel a workflow so it is no longer automatically retried or restarted"
)
def cancel(
    workflow_id: Annotated[str, typer.Argument()],
    application_database_url: Annotated[
        typing.Optional[str],
        typer.Option(
            "--db-url",
            "-D",
            help="Your DBOS application database URL",
        ),
    ] = None,
    system_database_url: Annotated[
        typing.Optional[str],
        typer.Option(
            "--sys-db-url",
            "-s",
            help="Your DBOS system database URL",
        ),
    ] = None,
) -> None:
    system_database_url, application_database_url = _get_db_url(
        system_database_url=system_database_url,
        application_database_url=application_database_url,
    )
    client = DBOSClient(
        application_database_url=application_database_url,
        system_database_url=system_database_url,
    )
    client.cancel_workflow(workflow_id=workflow_id)


@workflow.command(help="Resume a workflow that has been cancelled")
def resume(
    workflow_id: Annotated[str, typer.Argument()],
    application_database_url: Annotated[
        typing.Optional[str],
        typer.Option(
            "--db-url",
            "-D",
            help="Your DBOS application database URL",
        ),
    ] = None,
    system_database_url: Annotated[
        typing.Optional[str],
        typer.Option(
            "--sys-db-url",
            "-s",
            help="Your DBOS system database URL",
        ),
    ] = None,
) -> None:
    system_database_url, application_database_url = _get_db_url(
        system_database_url=system_database_url,
        application_database_url=application_database_url,
    )
    client = DBOSClient(
        application_database_url=application_database_url,
        system_database_url=system_database_url,
    )
    client.resume_workflow(workflow_id=workflow_id)


@workflow.command(
    help="[DEPRECATED - Use fork instead] Restart a workflow from the beginning with a new id"
)
def restart(
    workflow_id: Annotated[str, typer.Argument()],
    application_database_url: Annotated[
        typing.Optional[str],
        typer.Option(
            "--db-url",
            "-D",
            help="Your DBOS application database URL",
        ),
    ] = None,
    system_database_url: Annotated[
        typing.Optional[str],
        typer.Option(
            "--sys-db-url",
            "-s",
            help="Your DBOS system database URL",
        ),
    ] = None,
) -> None:
    system_database_url, application_database_url = _get_db_url(
        system_database_url=system_database_url,
        application_database_url=application_database_url,
    )
    client = DBOSClient(
        application_database_url=application_database_url,
        system_database_url=system_database_url,
    )
    status = client.fork_workflow(workflow_id=workflow_id, start_step=1).get_status()
    print(jsonpickle.encode(status, unpicklable=False))


@workflow.command(
    help="fork a workflow from the beginning with a new id and from a step"
)
def fork(
    workflow_id: Annotated[str, typer.Argument()],
    step: Annotated[
        int,
        typer.Option(
            "--step",
            "-S",
            help="Restart from this step",
        ),
    ] = 1,
    forked_workflow_id: Annotated[
        typing.Optional[str],
        typer.Option(
            "--forked-workflow-id",
            "-f",
            help="Custom ID for the forked workflow",
        ),
    ] = None,
    application_version: Annotated[
        typing.Optional[str],
        typer.Option(
            "--application-version",
            "-v",
            help="Custom application version for the forked workflow",
        ),
    ] = None,
    application_database_url: Annotated[
        typing.Optional[str],
        typer.Option(
            "--db-url",
            "-D",
            help="Your DBOS application database URL",
        ),
    ] = None,
    system_database_url: Annotated[
        typing.Optional[str],
        typer.Option(
            "--sys-db-url",
            "-s",
            help="Your DBOS system database URL",
        ),
    ] = None,
) -> None:
    system_database_url, application_database_url = _get_db_url(
        system_database_url=system_database_url,
        application_database_url=application_database_url,
    )
    client = DBOSClient(
        application_database_url=application_database_url,
        system_database_url=system_database_url,
    )

    if forked_workflow_id is not None:
        with SetWorkflowID(forked_workflow_id):
            status = client.fork_workflow(
                workflow_id=workflow_id,
                start_step=step,
                application_version=application_version,
            ).get_status()
    else:
        status = client.fork_workflow(
            workflow_id=workflow_id,
            start_step=step,
            application_version=application_version,
        ).get_status()
    print(jsonpickle.encode(status, unpicklable=False))


@queue.command(name="list", help="List enqueued functions for your application")
def list_queue(
    application_database_url: Annotated[
        typing.Optional[str],
        typer.Option(
            "--db-url",
            "-D",
            help="Your DBOS application database URL",
        ),
    ] = None,
    system_database_url: Annotated[
        typing.Optional[str],
        typer.Option(
            "--sys-db-url",
            "-s",
            help="Your DBOS system database URL",
        ),
    ] = None,
    limit: Annotated[
        typing.Optional[int],
        typer.Option("--limit", "-l", help="Limit the results returned"),
    ] = None,
    start_time: Annotated[
        typing.Optional[str],
        typer.Option(
            "--start-time",
            "-t",
            help="Retrieve functions starting after this timestamp (ISO 8601 format)",
        ),
    ] = None,
    end_time: Annotated[
        typing.Optional[str],
        typer.Option(
            "--end-time",
            "-e",
            help="Retrieve functions starting before this timestamp (ISO 8601 format)",
        ),
    ] = None,
    status: Annotated[
        typing.Optional[str],
        typer.Option(
            "--status",
            "-S",
            help="Retrieve functions with this status (PENDING, SUCCESS, ERROR, ENQUEUED, CANCELLED, or MAX_RECOVERY_ATTEMPTS_EXCEEDED)",
        ),
    ] = None,
    queue_name: Annotated[
        typing.Optional[str],
        typer.Option(
            "--queue-name",
            "-q",
            help="Retrieve functions on this queue",
        ),
    ] = None,
    name: Annotated[
        typing.Optional[str],
        typer.Option(
            "--name",
            "-n",
            help="Retrieve functions on this queue",
        ),
    ] = None,
    sort_desc: Annotated[
        bool,
        typer.Option(
            "--sort-desc",
            "-d",
            help="Sort the results in descending order (older first)",
        ),
    ] = False,
    offset: Annotated[
        typing.Optional[int],
        typer.Option(
            "--offset",
            "-o",
            help="Offset for pagination",
        ),
    ] = None,
) -> None:
    system_database_url, application_database_url = _get_db_url(
        system_database_url=system_database_url,
        application_database_url=application_database_url,
    )
    client = DBOSClient(
        application_database_url=application_database_url,
        system_database_url=system_database_url,
    )
    workflows = client.list_queued_workflows(
        limit=limit,
        offset=offset,
        sort_desc=sort_desc,
        start_time=start_time,
        end_time=end_time,
        queue_name=queue_name,
        status=status,
        name=name,
    )
    print(jsonpickle.encode(workflows, unpicklable=False))


if __name__ == "__main__":
    app()



import os
import tempfile
import unittest
from typing import Any, Dict, List, TypedDict, cast
from unittest.mock import mock_open, patch

from sqlalchemy import make_url

from dbos._dbos_config import _substitute_env_vars, load_config


class ConfigFile(TypedDict, total=False):
    name: str
    database: Dict[str, Any]
    database_url: str
    telemetry: Dict[str, Any]
    runtimeConfig: Dict[str, List[str]]


class DatabaseConfig(TypedDict, total=False):
    hostname: str
    port: int
    username: str
    password: str
    app_db_name: str
    url: str
    nested: Dict[str, Any]
    # Add other fields as needed


class TestDockerSecrets(unittest.TestCase):
    def setUp(self) -> None:
        # Create a temporary directory to simulate /run/secrets/
        self.temp_dir = tempfile.TemporaryDirectory()
        self.secrets_dir = os.path.join(self.temp_dir.name, "secrets")
        os.makedirs(self.secrets_dir)

    def tearDown(self) -> None:
        self.temp_dir.cleanup()

    def test_substitute_env_vars_with_env_vars(self) -> None:
        # Test that environment variables are still substituted correctly
        with patch.dict(os.environ, {"TEST_VAR": "test_value"}):
            content = "This is a ${TEST_VAR} test"
            result = _substitute_env_vars(content)
            self.assertEqual(result, "This is a test_value test")

    def test_substitute_env_vars_with_docker_secrets(self) -> None:
        # Create a mock Docker secret
        secret_path = os.path.join(self.secrets_dir, "db_password")
        with open(secret_path, "w") as f:
            f.write("secret_password")

        # Mock the /run/secrets/ path
        with patch("os.path.exists") as mock_exists:
            mock_exists.return_value = True
            with patch("builtins.open", mock_open(read_data="secret_password")):
                content = "This is a ${DOCKER_SECRET:db_password} test"
                result = _substitute_env_vars(content)
                self.assertEqual(result, "This is a secret_password test")

    def test_substitute_env_vars_with_missing_docker_secret(self) -> None:
        # Test that a warning is logged when a Docker secret is missing
        with patch("dbos._dbos_config.dbos_logger") as mock_logger:
            content = "This is a ${DOCKER_SECRET:missing_secret} test"
            result = _substitute_env_vars(content)
            self.assertEqual(result, "This is a  test")
            mock_logger.warning.assert_called_once()

    def test_substitute_env_vars_with_both_env_vars_and_docker_secrets(self) -> None:
        # Test that both environment variables and Docker secrets are substituted
        with patch.dict(os.environ, {"TEST_VAR": "test_value"}):
            with patch("os.path.exists") as mock_exists:
                mock_exists.return_value = True
                with patch(
                    "builtins.open",
                    mock_open(read_data="secret_password"),
                ):
                    content = (
                        "This is a ${TEST_VAR} and ${DOCKER_SECRET:db_password} test"
                    )
                    result = _substitute_env_vars(content)
                    self.assertEqual(
                        result, "This is a test_value and secret_password test"
                    )

    def test_substitute_env_vars_with_silent_mode(self) -> None:
        # Test that no warning is logged when silent mode is enabled
        with patch("dbos._dbos_config.dbos_logger") as mock_logger:
            content = "This is a ${DOCKER_SECRET:missing_secret} test"
            result = _substitute_env_vars(content, silent=True)
            self.assertEqual(result, "This is a  test")
            mock_logger.warning.assert_not_called()

    def test_load_config_with_docker_secrets_in_database_url(self) -> None:
        # Create a mock configuration file with Docker secrets in the database URL
        config_content = """
name: test-app
database_url: postgresql://postgres:${DOCKER_SECRET:db_password}@localhost:5432/test_db
"""

        # Mock the file open and read operations
        mock_file = mock_open(read_data=config_content)

        # Create a mock dictionary that would be returned by yaml.safe_load
        mock_config_dict: ConfigFile = {
            "name": "test-app",
            "database_url": "postgresql://postgres:secret_password@localhost:5432/test_db",
        }

        # Mock the schema validation to always pass
        with (
            patch("builtins.open", mock_file),
            patch("dbos._dbos_config.validate") as mock_validate,
            patch("dbos._dbos_config.resources.files") as mock_resources,
            patch("os.path.exists") as mock_exists,
            patch("yaml.safe_load") as mock_yaml_load,
        ):

            # Set up the mocks
            mock_exists.return_value = True
            mock_resources.return_value.joinpath.return_value.open.return_value.__enter__.return_value.read.return_value = (
                "{}"
            )
            mock_yaml_load.return_value = mock_config_dict

            # Call the load_config function
            config = load_config()

            # Verify that the Docker secret was correctly substituted in the database URL
            self.assertEqual(
                config["database_url"],
                "postgresql://postgres:secret_password@localhost:5432/test_db",
            )

            # Verify that the schema validation was called
            mock_validate.assert_called_once()

    def test_load_config_with_docker_secrets_in_list(self) -> None:
        # Create a mock configuration file with Docker secrets in a list
        config_content = """
name: test-app
runtimeConfig:
  setup:
    - echo "Setting up environment"
    - export API_KEY=${DOCKER_SECRET:api_key}
    - export DB_PASSWORD=${DOCKER_SECRET:db_password}
"""

        # Mock the file open and read operations
        mock_file = mock_open(read_data=config_content)

        # Create a mock dictionary that would be returned by yaml.safe_load
        mock_config_dict: ConfigFile = {
            "name": "test-app",
            "runtimeConfig": {
                "setup": [
                    'echo "Setting up environment"',
                    "export API_KEY=secret_value",
                    "export DB_PASSWORD=secret_value",
                ]
            },
        }

        # Mock the schema validation to always pass
        with (
            patch("builtins.open", mock_file),
            patch("dbos._dbos_config.validate") as mock_validate,
            patch("dbos._dbos_config.resources.files") as mock_resources,
            patch("os.path.exists") as mock_exists,
            patch("yaml.safe_load") as mock_yaml_load,
        ):

            # Set up the mocks
            mock_exists.return_value = True
            mock_resources.return_value.joinpath.return_value.open.return_value.__enter__.return_value.read.return_value = (
                "{}"
            )
            mock_yaml_load.return_value = mock_config_dict

            # Call the load_config function
            config = load_config()

            # Verify that the Docker secrets were correctly substituted in the list
            setup_commands = config["runtimeConfig"]["setup"]
            if setup_commands is not None:
                self.assertEqual(setup_commands[0], 'echo "Setting up environment"')
                self.assertEqual(setup_commands[1], "export API_KEY=secret_value")
                self.assertEqual(setup_commands[2], "export DB_PASSWORD=secret_value")

            # Verify that the schema validation was called
            mock_validate.assert_called_once()

    def test_load_config_with_multiple_docker_secrets_in_string(self) -> None:
        # Create a mock configuration file with multiple Docker secrets in a string
        config_content = """
name: test-app
database_url: postgresql://${DOCKER_SECRET:db_user}:${DOCKER_SECRET:db_password}@${DOCKER_SECRET:db_host}:${DOCKER_SECRET:db_port}/${DOCKER_SECRET:db_name}
"""

        # Mock the file open and read operations
        mock_file = mock_open(read_data=config_content)

        # Create a mock dictionary that would be returned by yaml.safe_load
        mock_config_dict: ConfigFile = {
            "name": "test-app",
            "database_url": "postgresql://user:pass@host:5432/db",
        }

        # Mock the schema validation to always pass
        with (
            patch("builtins.open", mock_file),
            patch("dbos._dbos_config.validate") as mock_validate,
            patch("dbos._dbos_config.resources.files") as mock_resources,
            patch("os.path.exists") as mock_exists,
            patch("yaml.safe_load") as mock_yaml_load,
        ):

            # Set up the mocks
            mock_exists.return_value = True
            mock_resources.return_value.joinpath.return_value.open.return_value.__enter__.return_value.read.return_value = (
                "{}"
            )
            mock_yaml_load.return_value = mock_config_dict

            def mock_secret_open(*args: Any, **kwargs: Any) -> Any:
                filepath = args[0]
                if filepath == "dbos-config.yaml":
                    return mock_open(read_data=config_content).return_value

                secret_name = filepath.split("/")[-1]
                secret_values = {
                    "db_user": "user",
                    "db_password": "pass",
                    "db_host": "host",
                    "db_port": "5432",
                    "db_name": "db",
                }
                if secret_name in secret_values:
                    return mock_open(read_data=secret_values[secret_name]).return_value
                raise FileNotFoundError(f"Mock file not found: {filepath}")

            with patch("builtins.open", mock_secret_open, create=True):
                # Call the load_config function
                config = load_config()

                # Verify that all Docker secrets were correctly substituted in the string
                self.assertEqual(
                    config["database_url"],
                    "postgresql://user:pass@host:5432/db",
                )

                # Verify that the schema validation was called
                mock_validate.assert_called_once()

    def test_load_config_without_docker_secrets(self) -> None:
        # Create a mock configuration file without Docker secrets
        config_content = """
name: test-app
database_url: postgresql://postgres:plain_password@localhost:5432/test_db
"""

        # Mock the file open and read operations
        mock_file = mock_open(read_data=config_content)

        # Create a mock dictionary that would be returned by yaml.safe_load
        mock_config_dict: ConfigFile = {
            "name": "test-app",
            "database_url": "postgresql://postgres:plain_password@localhost:5432/test_db",
        }

        # Mock the schema validation to always pass
        with (
            patch("builtins.open", mock_file),
            patch("dbos._dbos_config.validate") as mock_validate,
            patch("dbos._dbos_config.resources.files") as mock_resources,
            patch("yaml.safe_load") as mock_yaml_load,
        ):

            # Set up the mocks
            mock_resources.return_value.joinpath.return_value.open.return_value.__enter__.return_value.read.return_value = (
                "{}"
            )
            mock_yaml_load.return_value = mock_config_dict

            # Call the load_config function
            config = load_config()

            # Verify that the configuration was loaded correctly without any substitutions
            assert config["database_url"] is not None
            self.assertEqual(
                make_url(config["database_url"]).password, "plain_password"
            )

            # Verify that the schema validation was called
            mock_validate.assert_called_once()

    def test_load_config_with_mixed_env_vars_and_docker_secrets(self) -> None:
        # Create a mock configuration file with both environment variables and Docker secrets
        config_content = """
name: test-app
database_url: postgresql://${DB_USER}:${DOCKER_SECRET:db_password}@${DB_HOST}:${DB_PORT}/${DB_NAME}
"""

        # Mock the file open and read operations
        mock_file = mock_open(read_data=config_content)

        # Create a mock dictionary that would be returned by yaml.safe_load
        mock_config_dict: ConfigFile = {
            "name": "test-app",
            "database_url": "postgresql://postgres:secret_password@localhost:5432/test_db",
        }

        # Mock the schema validation to always pass
        with (
            patch("builtins.open", mock_file),
            patch("dbos._dbos_config.validate") as mock_validate,
            patch("dbos._dbos_config.resources.files") as mock_resources,
            patch("os.path.exists") as mock_exists,
            patch(
                "builtins.open", mock_open(read_data="secret_password"), create=True
            ) as mock_secret_file,
            patch("yaml.safe_load") as mock_yaml_load,
            patch.dict(
                os.environ,
                {
                    "DB_HOST": "localhost",
                    "DB_PORT": "5432",
                    "DB_USER": "postgres",
                    "DB_NAME": "test_db",
                },
            ),
        ):

            # Set up the mocks
            mock_exists.return_value = True
            mock_resources.return_value.joinpath.return_value.open.return_value.__enter__.return_value.read.return_value = (
                "{}"
            )
            mock_yaml_load.return_value = mock_config_dict

            # Call the load_config function
            config = load_config()

            # Verify that both environment variables and Docker secrets were correctly substituted
            self.assertEqual(
                config["database_url"],
                "postgresql://postgres:secret_password@localhost:5432/test_db",
            )

            # Verify that the schema validation was called
            mock_validate.assert_called_once()


if __name__ == "__main__":
    unittest.main()



from dbos import DBOS
from dbos._dbos_config import DBOSConfig


@DBOS.workflow()
def my_function(foo: str) -> str:
    return foo


def default_config() -> DBOSConfig:
    return {
        "name": "forgot-launch",
        "database_url": f"postgresql://postgres:doesntmatter@localhost:5432/notneeded",
    }


DBOS(config=default_config())



"""consolidate_queues

Revision ID: 66478e1b95e5
Revises: 933e86bdac6a
Create Date: 2025-05-21 10:14:25.674613

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "66478e1b95e5"
down_revision: Union[str, None] = "933e86bdac6a"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # Add new columns to workflow_status table
    op.add_column(
        "workflow_status",
        sa.Column("started_at_epoch_ms", sa.BigInteger(), nullable=True),
        schema="dbos",
    )

    op.add_column(
        "workflow_status",
        sa.Column("deduplication_id", sa.Text(), nullable=True),
        schema="dbos",
    )

    op.add_column(
        "workflow_status",
        sa.Column(
            "priority", sa.Integer(), nullable=False, server_default=sa.text("'0'::int")
        ),
        schema="dbos",
    )

    # Add unique constraint for deduplication_id
    op.create_unique_constraint(
        "uq_workflow_status_queue_name_dedup_id",
        "workflow_status",
        ["queue_name", "deduplication_id"],
        schema="dbos",
    )

    # Add index on status field
    op.create_index(
        "workflow_status_status_index", "workflow_status", ["status"], schema="dbos"
    )


def downgrade() -> None:
    # Drop indexes
    op.drop_index(
        "workflow_status_status_index", table_name="workflow_status", schema="dbos"
    )

    # Drop unique constraint
    op.drop_constraint(
        "uq_workflow_status_queue_name_dedup_id", "workflow_status", schema="dbos"
    )

    # Drop columns
    op.drop_column("workflow_status", "priority", schema="dbos")
    op.drop_column("workflow_status", "deduplication_id", schema="dbos")
    op.drop_column("workflow_status", "started_at_epoch_ms", schema="dbos")



import asyncio
import sys
import time
import uuid
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    Generator,
    Generic,
    List,
    Optional,
    TypedDict,
    TypeVar,
    Union,
)

from dbos._app_db import ApplicationDatabase
from dbos._context import MaxPriority, MinPriority
from dbos._sys_db import SystemDatabase

if sys.version_info < (3, 11):
    from typing_extensions import NotRequired
else:
    from typing import NotRequired

from dbos import _serialization

if TYPE_CHECKING:
    from dbos._dbos import WorkflowHandle, WorkflowHandleAsync

from dbos._dbos_config import (
    get_application_database_url,
    get_system_database_url,
    is_valid_database_url,
)
from dbos._error import DBOSException, DBOSNonExistentWorkflowError
from dbos._registrations import DEFAULT_MAX_RECOVERY_ATTEMPTS
from dbos._serialization import WorkflowInputs
from dbos._sys_db import (
    EnqueueOptionsInternal,
    StepInfo,
    SystemDatabase,
    WorkflowStatus,
    WorkflowStatusInternal,
    WorkflowStatusString,
    _dbos_stream_closed_sentinel,
    workflow_is_active,
)
from dbos._workflow_commands import (
    fork_workflow,
    get_workflow,
    list_queued_workflows,
    list_workflow_steps,
    list_workflows,
)

R = TypeVar("R", covariant=True)  # A generic type for workflow return values


class EnqueueOptions(TypedDict):
    workflow_name: str
    queue_name: str
    workflow_id: NotRequired[str]
    app_version: NotRequired[str]
    workflow_timeout: NotRequired[float]
    deduplication_id: NotRequired[str]
    priority: NotRequired[int]


def validate_enqueue_options(options: EnqueueOptions) -> None:
    priority = options.get("priority")
    if priority is not None and (priority < MinPriority or priority > MaxPriority):
        raise DBOSException(
            f"Invalid priority {priority}. Priority must be between {MinPriority}~{MaxPriority}."
        )


class WorkflowHandleClientPolling(Generic[R]):

    def __init__(self, workflow_id: str, sys_db: SystemDatabase):
        self.workflow_id = workflow_id
        self._sys_db = sys_db

    def get_workflow_id(self) -> str:
        return self.workflow_id

    def get_result(self) -> R:
        res: R = self._sys_db.await_workflow_result(self.workflow_id)
        return res

    def get_status(self) -> WorkflowStatus:
        status = get_workflow(self._sys_db, self.workflow_id)
        if status is None:
            raise DBOSNonExistentWorkflowError(self.workflow_id)
        return status


class WorkflowHandleClientAsyncPolling(Generic[R]):

    def __init__(self, workflow_id: str, sys_db: SystemDatabase):
        self.workflow_id = workflow_id
        self._sys_db = sys_db

    def get_workflow_id(self) -> str:
        return self.workflow_id

    async def get_result(self) -> R:
        res: R = await asyncio.to_thread(
            self._sys_db.await_workflow_result, self.workflow_id
        )
        return res

    async def get_status(self) -> WorkflowStatus:
        status = await asyncio.to_thread(get_workflow, self._sys_db, self.workflow_id)
        if status is None:
            raise DBOSNonExistentWorkflowError(self.workflow_id)
        return status


class DBOSClient:
    def __init__(
        self,
        database_url: Optional[str] = None,  # DEPRECATED
        *,
        system_database_url: Optional[str] = None,
        application_database_url: Optional[str] = None,
        system_database: Optional[str] = None,  # DEPRECATED
    ):
        application_database_url = get_application_database_url(
            {
                "system_database_url": system_database_url,
                "database_url": (
                    database_url if database_url else application_database_url
                ),
            }
        )
        system_database_url = get_system_database_url(
            {
                "system_database_url": system_database_url,
                "database_url": application_database_url,
                "database": {"sys_db_name": system_database},
            }
        )
        assert is_valid_database_url(system_database_url)
        assert is_valid_database_url(application_database_url)
        # We only create database connections but do not run migrations
        self._sys_db = SystemDatabase.create(
            system_database_url=system_database_url,
            engine_kwargs={
                "pool_timeout": 30,
                "max_overflow": 0,
                "pool_size": 2,
            },
        )
        self._sys_db.check_connection()
        self._app_db = ApplicationDatabase.create(
            database_url=application_database_url,
            engine_kwargs={
                "pool_timeout": 30,
                "max_overflow": 0,
                "pool_size": 2,
            },
        )

    def destroy(self) -> None:
        self._sys_db.destroy()

    def _enqueue(self, options: EnqueueOptions, *args: Any, **kwargs: Any) -> str:
        validate_enqueue_options(options)
        workflow_name = options["workflow_name"]
        queue_name = options["queue_name"]

        max_recovery_attempts = options.get("max_recovery_attempts")
        if max_recovery_attempts is None:
            max_recovery_attempts = DEFAULT_MAX_RECOVERY_ATTEMPTS
        workflow_id = options.get("workflow_id")
        if workflow_id is None:
            workflow_id = str(uuid.uuid4())
        workflow_timeout = options.get("workflow_timeout", None)
        enqueue_options_internal: EnqueueOptionsInternal = {
            "deduplication_id": options.get("deduplication_id"),
            "priority": options.get("priority"),
            "app_version": options.get("app_version"),
        }

        inputs: WorkflowInputs = {
            "args": args,
            "kwargs": kwargs,
        }

        status: WorkflowStatusInternal = {
            "workflow_uuid": workflow_id,
            "status": WorkflowStatusString.ENQUEUED.value,
            "name": workflow_name,
            "class_name": None,
            "queue_name": queue_name,
            "app_version": enqueue_options_internal["app_version"],
            "config_name": None,
            "authenticated_user": None,
            "assumed_role": None,
            "authenticated_roles": None,
            "output": None,
            "error": None,
            "created_at": None,
            "updated_at": None,
            "executor_id": None,
            "recovery_attempts": None,
            "app_id": None,
            "workflow_timeout_ms": (
                int(workflow_timeout * 1000) if workflow_timeout is not None else None
            ),
            "workflow_deadline_epoch_ms": None,
            "deduplication_id": enqueue_options_internal["deduplication_id"],
            "priority": (
                enqueue_options_internal["priority"]
                if enqueue_options_internal["priority"] is not None
                else 0
            ),
            "inputs": _serialization.serialize_args(inputs),
        }

        self._sys_db.init_workflow(
            status,
            max_recovery_attempts=None,
        )
        return workflow_id

    def enqueue(
        self, options: EnqueueOptions, *args: Any, **kwargs: Any
    ) -> "WorkflowHandle[R]":
        workflow_id = self._enqueue(options, *args, **kwargs)
        return WorkflowHandleClientPolling[R](workflow_id, self._sys_db)

    async def enqueue_async(
        self, options: EnqueueOptions, *args: Any, **kwargs: Any
    ) -> "WorkflowHandleAsync[R]":
        workflow_id = await asyncio.to_thread(self._enqueue, options, *args, **kwargs)
        return WorkflowHandleClientAsyncPolling[R](workflow_id, self._sys_db)

    def retrieve_workflow(self, workflow_id: str) -> "WorkflowHandle[R]":
        status = get_workflow(self._sys_db, workflow_id)
        if status is None:
            raise DBOSNonExistentWorkflowError(workflow_id)
        return WorkflowHandleClientPolling[R](workflow_id, self._sys_db)

    async def retrieve_workflow_async(
        self, workflow_id: str
    ) -> "WorkflowHandleAsync[R]":
        status = await asyncio.to_thread(get_workflow, self._sys_db, workflow_id)
        if status is None:
            raise DBOSNonExistentWorkflowError(workflow_id)
        return WorkflowHandleClientAsyncPolling[R](workflow_id, self._sys_db)

    def send(
        self,
        destination_id: str,
        message: Any,
        topic: Optional[str] = None,
        idempotency_key: Optional[str] = None,
    ) -> None:
        idempotency_key = idempotency_key if idempotency_key else str(uuid.uuid4())
        status: WorkflowStatusInternal = {
            "workflow_uuid": f"{destination_id}-{idempotency_key}",
            "status": WorkflowStatusString.SUCCESS.value,
            "name": "temp_workflow-send-client",
            "class_name": None,
            "queue_name": None,
            "config_name": None,
            "authenticated_user": None,
            "assumed_role": None,
            "authenticated_roles": None,
            "output": None,
            "error": None,
            "created_at": None,
            "updated_at": None,
            "executor_id": None,
            "recovery_attempts": None,
            "app_id": None,
            "app_version": None,
            "workflow_timeout_ms": None,
            "workflow_deadline_epoch_ms": None,
            "deduplication_id": None,
            "priority": 0,
            "inputs": _serialization.serialize_args({"args": (), "kwargs": {}}),
        }
        with self._sys_db.engine.begin() as conn:
            self._sys_db._insert_workflow_status(
                status, conn, max_recovery_attempts=None
            )
        self._sys_db.send(status["workflow_uuid"], 0, destination_id, message, topic)

    async def send_async(
        self,
        destination_id: str,
        message: Any,
        topic: Optional[str] = None,
        idempotency_key: Optional[str] = None,
    ) -> None:
        return await asyncio.to_thread(
            self.send, destination_id, message, topic, idempotency_key
        )

    def get_event(self, workflow_id: str, key: str, timeout_seconds: float = 60) -> Any:
        return self._sys_db.get_event(workflow_id, key, timeout_seconds)

    async def get_event_async(
        self, workflow_id: str, key: str, timeout_seconds: float = 60
    ) -> Any:
        return await asyncio.to_thread(
            self.get_event, workflow_id, key, timeout_seconds
        )

    def cancel_workflow(self, workflow_id: str) -> None:
        self._sys_db.cancel_workflow(workflow_id)

    async def cancel_workflow_async(self, workflow_id: str) -> None:
        await asyncio.to_thread(self.cancel_workflow, workflow_id)

    def resume_workflow(self, workflow_id: str) -> "WorkflowHandle[Any]":
        self._sys_db.resume_workflow(workflow_id)
        return WorkflowHandleClientPolling[Any](workflow_id, self._sys_db)

    async def resume_workflow_async(
        self, workflow_id: str
    ) -> "WorkflowHandleAsync[Any]":
        await asyncio.to_thread(self.resume_workflow, workflow_id)
        return WorkflowHandleClientAsyncPolling[Any](workflow_id, self._sys_db)

    def list_workflows(
        self,
        *,
        workflow_ids: Optional[List[str]] = None,
        status: Optional[Union[str, List[str]]] = None,
        start_time: Optional[str] = None,
        end_time: Optional[str] = None,
        name: Optional[str] = None,
        app_version: Optional[str] = None,
        user: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        sort_desc: bool = False,
        workflow_id_prefix: Optional[str] = None,
        load_input: bool = True,
        load_output: bool = True,
    ) -> List[WorkflowStatus]:
        return list_workflows(
            self._sys_db,
            workflow_ids=workflow_ids,
            status=status,
            start_time=start_time,
            end_time=end_time,
            name=name,
            app_version=app_version,
            user=user,
            limit=limit,
            offset=offset,
            sort_desc=sort_desc,
            workflow_id_prefix=workflow_id_prefix,
            load_input=load_input,
            load_output=load_output,
        )

    async def list_workflows_async(
        self,
        *,
        workflow_ids: Optional[List[str]] = None,
        status: Optional[Union[str, List[str]]] = None,
        start_time: Optional[str] = None,
        end_time: Optional[str] = None,
        name: Optional[str] = None,
        app_version: Optional[str] = None,
        user: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        sort_desc: bool = False,
        workflow_id_prefix: Optional[str] = None,
        load_input: bool = True,
        load_output: bool = True,
    ) -> List[WorkflowStatus]:
        return await asyncio.to_thread(
            self.list_workflows,
            workflow_ids=workflow_ids,
            status=status,
            start_time=start_time,
            end_time=end_time,
            name=name,
            app_version=app_version,
            user=user,
            limit=limit,
            offset=offset,
            sort_desc=sort_desc,
            workflow_id_prefix=workflow_id_prefix,
            load_input=load_input,
            load_output=load_output,
        )

    def list_queued_workflows(
        self,
        *,
        queue_name: Optional[str] = None,
        status: Optional[Union[str, List[str]]] = None,
        start_time: Optional[str] = None,
        end_time: Optional[str] = None,
        name: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        sort_desc: bool = False,
        load_input: bool = True,
    ) -> List[WorkflowStatus]:
        return list_queued_workflows(
            self._sys_db,
            queue_name=queue_name,
            status=status,
            start_time=start_time,
            end_time=end_time,
            name=name,
            limit=limit,
            offset=offset,
            sort_desc=sort_desc,
            load_input=load_input,
        )

    async def list_queued_workflows_async(
        self,
        *,
        queue_name: Optional[str] = None,
        status: Optional[Union[str, List[str]]] = None,
        start_time: Optional[str] = None,
        end_time: Optional[str] = None,
        name: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        sort_desc: bool = False,
        load_input: bool = True,
    ) -> List[WorkflowStatus]:
        return await asyncio.to_thread(
            self.list_queued_workflows,
            queue_name=queue_name,
            status=status,
            start_time=start_time,
            end_time=end_time,
            name=name,
            limit=limit,
            offset=offset,
            sort_desc=sort_desc,
            load_input=load_input,
        )

    def list_workflow_steps(self, workflow_id: str) -> List[StepInfo]:
        return list_workflow_steps(self._sys_db, self._app_db, workflow_id)

    async def list_workflow_steps_async(self, workflow_id: str) -> List[StepInfo]:
        return await asyncio.to_thread(self.list_workflow_steps, workflow_id)

    def fork_workflow(
        self,
        workflow_id: str,
        start_step: int,
        *,
        application_version: Optional[str] = None,
    ) -> "WorkflowHandle[Any]":
        forked_workflow_id = fork_workflow(
            self._sys_db,
            self._app_db,
            workflow_id,
            start_step,
            application_version=application_version,
        )
        return WorkflowHandleClientPolling[Any](forked_workflow_id, self._sys_db)

    async def fork_workflow_async(
        self,
        workflow_id: str,
        start_step: int,
        *,
        application_version: Optional[str] = None,
    ) -> "WorkflowHandleAsync[Any]":
        forked_workflow_id = await asyncio.to_thread(
            fork_workflow,
            self._sys_db,
            self._app_db,
            workflow_id,
            start_step,
            application_version=application_version,
        )
        return WorkflowHandleClientAsyncPolling[Any](forked_workflow_id, self._sys_db)

    def read_stream(self, workflow_id: str, key: str) -> Generator[Any, Any, None]:
        """
        Read values from a stream as a generator.
        This function reads values from a stream identified by the workflow_id and key,
        yielding each value in order until the stream is closed or the workflow terminates.

        Args:
            workflow_id: The ID of the workflow that wrote to the stream
            key: The stream key to read from

        Yields:
            The values written to the stream in order
        """
        offset = 0
        while True:
            try:
                value = self._sys_db.read_stream(workflow_id, key, offset)
                if value == _dbos_stream_closed_sentinel:
                    break
                yield value
                offset += 1
            except ValueError:
                # Poll the offset until a value arrives or the workflow terminates
                status = get_workflow(self._sys_db, workflow_id)
                if status is None:
                    break
                if not workflow_is_active(status.status):
                    break
                time.sleep(1.0)
                continue

    async def read_stream_async(
        self, workflow_id: str, key: str
    ) -> AsyncGenerator[Any, None]:
        """
        Read values from a stream as an async generator.
        This function reads values from a stream identified by the workflow_id and key,
        yielding each value in order until the stream is closed or the workflow terminates.

        Args:
            workflow_id: The ID of the workflow that wrote to the stream
            key: The stream key to read from

        Yields:
            The values written to the stream in order
        """
        offset = 0
        while True:
            try:
                value = await asyncio.to_thread(
                    self._sys_db.read_stream, workflow_id, key, offset
                )
                if value == _dbos_stream_closed_sentinel:
                    break
                yield value
                offset += 1
            except ValueError:
                # Poll the offset until a value arrives or the workflow terminates
                status = await asyncio.to_thread(
                    get_workflow, self._sys_db, workflow_id
                )
                if status is None:
                    break
                if not workflow_is_active(status.status):
                    break
                await asyncio.sleep(1.0)
                continue



import sqlalchemy as sa
import typer

from dbos._app_db import ApplicationDatabase
from dbos._sys_db import SystemDatabase


def migrate_dbos_databases(app_database_url: str, system_database_url: str) -> None:
    app_db = None
    sys_db = None
    try:
        sys_db = SystemDatabase.create(
            system_database_url=system_database_url,
            engine_kwargs={
                "pool_timeout": 30,
                "max_overflow": 0,
                "pool_size": 2,
            },
        )
        app_db = ApplicationDatabase.create(
            database_url=app_database_url,
            engine_kwargs={
                "pool_timeout": 30,
                "max_overflow": 0,
                "pool_size": 2,
            },
        )
        sys_db.run_migrations()
        app_db.run_migrations()
    except Exception as e:
        typer.echo(f"DBOS migrations failed: {e}")
        raise typer.Exit(code=1)
    finally:
        if sys_db:
            sys_db.destroy()
        if app_db:
            app_db.destroy()


def grant_dbos_schema_permissions(database_url: str, role_name: str) -> None:
    """
    Grant all permissions on all entities in the dbos schema to the specified role.
    """
    typer.echo(
        f"Granting permissions for DBOS schema to {role_name} in database {sa.make_url(database_url)}"
    )
    engine = None
    try:
        engine = sa.create_engine(
            sa.make_url(database_url).set(drivername="postgresql+psycopg")
        )
        with engine.connect() as connection:
            connection.execution_options(isolation_level="AUTOCOMMIT")

            # Grant usage on the dbos schema
            sql = f'GRANT USAGE ON SCHEMA dbos TO "{role_name}"'
            typer.echo(sql)
            connection.execute(sa.text(sql))

            # Grant all privileges on all existing tables in dbos schema (includes views)
            sql = f'GRANT ALL PRIVILEGES ON ALL TABLES IN SCHEMA dbos TO "{role_name}"'
            typer.echo(sql)
            connection.execute(sa.text(sql))

            # Grant all privileges on all sequences in dbos schema
            sql = (
                f'GRANT ALL PRIVILEGES ON ALL SEQUENCES IN SCHEMA dbos TO "{role_name}"'
            )
            typer.echo(sql)
            connection.execute(sa.text(sql))

            # Grant execute on all functions and procedures in dbos schema
            sql = f'GRANT EXECUTE ON ALL FUNCTIONS IN SCHEMA dbos TO "{role_name}"'
            typer.echo(sql)
            connection.execute(sa.text(sql))

            # Grant default privileges for future objects in dbos schema
            sql = f'ALTER DEFAULT PRIVILEGES IN SCHEMA dbos GRANT ALL ON TABLES TO "{role_name}"'
            typer.echo(sql)
            connection.execute(sa.text(sql))

            sql = f'ALTER DEFAULT PRIVILEGES IN SCHEMA dbos GRANT ALL ON SEQUENCES TO "{role_name}"'
            typer.echo(sql)
            connection.execute(sa.text(sql))

            sql = f'ALTER DEFAULT PRIVILEGES IN SCHEMA dbos GRANT EXECUTE ON FUNCTIONS TO "{role_name}"'
            typer.echo(sql)
            connection.execute(sa.text(sql))

    except Exception as e:
        typer.echo(f"Failed to grant permissions to role {role_name}: {e}")
        raise typer.Exit(code=1)
    finally:
        if engine:
            engine.dispose()



import importlib
import math
import os
import runpy
import subprocess
import sys
import time
import uuid
from typing import Optional, TypedDict

import pytest
import sqlalchemy as sa
from sqlalchemy.exc import DBAPIError

from dbos import DBOS, DBOSClient, DBOSConfig, EnqueueOptions, SetWorkflowID
from dbos._dbos import WorkflowHandle, WorkflowHandleAsync
from dbos._sys_db import SystemDatabase
from dbos._utils import GlobalParams
from tests import client_collateral
from tests.client_collateral import event_test, retrieve_test, send_test


class Person(TypedDict):
    first: str
    last: str
    age: int


def run_client_collateral() -> None:
    dirname = os.path.dirname(__file__)
    filename = os.path.join(dirname, "client_collateral.py")
    runpy.run_path(filename)


def test_client_no_migrate(
    dbos: DBOS, config: DBOSConfig, skip_with_sqlite: None
) -> None:
    # Drop the system database
    DBOS.destroy()
    DBOS(config=config)
    DBOS.reset_system_database()

    # The client should not be able to connect to the system database
    with pytest.raises(Exception) as exc_info:
        assert config["application_database_url"]
        client = DBOSClient(config["application_database_url"])
        client.list_workflows()
    assert f'database "dbostestpy_dbos_sys" does not exist' in str(exc_info.value)


def test_client_enqueue_and_get_result(dbos: DBOS, client: DBOSClient) -> None:
    run_client_collateral()

    johnDoe: Person = {"first": "John", "last": "Doe", "age": 30}
    wfid = str(uuid.uuid4())

    options: EnqueueOptions = {
        "queue_name": "test_queue",
        "workflow_name": "enqueue_test",
        "workflow_id": wfid,
    }

    handle: WorkflowHandle[str] = client.enqueue(options, 42, "test", johnDoe)
    result = handle.get_result()
    assert result == '42-test-{"first": "John", "last": "Doe", "age": 30}'

    list_results = client.list_workflows()
    assert len(list_results) == 1
    assert list_results[0].workflow_id == wfid
    assert list_results[0].status == "SUCCESS"
    assert list_results[0].output == result
    assert list_results[0].input is not None

    # Skip loading input and output
    list_results = client.list_workflows(load_input=False, load_output=False)
    assert len(list_results) == 1
    assert list_results[0].workflow_id == wfid
    assert list_results[0].status == "SUCCESS"
    assert list_results[0].output is None
    assert list_results[0].input is None


def test_enqueue_with_timeout(dbos: DBOS, client: DBOSClient) -> None:
    run_client_collateral()

    wfid = str(uuid.uuid4())
    options: EnqueueOptions = {
        "queue_name": "test_queue",
        "workflow_name": "blocked_workflow",
        "workflow_timeout": 1,
        "workflow_id": wfid,
    }

    handle: WorkflowHandle[str] = client.enqueue(options)

    list_results = client.list_queued_workflows()
    assert len(list_results) == 1
    assert list_results[0].workflow_id == wfid
    assert list_results[0].status in ["PENDING", "ENQUEUED"]

    with pytest.raises(Exception) as exc_info:
        handle.get_result()
    assert "was cancelled" in str(exc_info.value)


def test_client_enqueue_appver_not_set(dbos: DBOS, client: DBOSClient) -> None:
    run_client_collateral()

    johnDoe: Person = {"first": "John", "last": "Doe", "age": 30}
    wfid = str(uuid.uuid4())

    options: EnqueueOptions = {
        "queue_name": "test_queue",
        "workflow_name": "enqueue_test",
        "workflow_id": wfid,
    }

    client.enqueue(options, 42, "test", johnDoe)

    handle: WorkflowHandle[str] = DBOS.retrieve_workflow(wfid)
    result = handle.get_result()
    assert result == '42-test-{"first": "John", "last": "Doe", "age": 30}'

    wf_status = dbos.get_workflow_status(wfid)
    assert wf_status is not None
    assert wf_status.status == "SUCCESS"
    assert wf_status.name == "enqueue_test"
    assert wf_status.app_version == GlobalParams.app_version


def test_client_enqueue_appver_set(dbos: DBOS, client: DBOSClient) -> None:
    run_client_collateral()

    johnDoe: Person = {"first": "John", "last": "Doe", "age": 30}
    wfid = str(uuid.uuid4())

    options: EnqueueOptions = {
        "queue_name": "test_queue",
        "workflow_name": "enqueue_test",
        "workflow_id": wfid,
        "app_version": GlobalParams.app_version,
    }

    client.enqueue(options, 42, "test", johnDoe)

    handle: WorkflowHandle[str] = DBOS.retrieve_workflow(wfid)
    result = handle.get_result()
    assert result == '42-test-{"first": "John", "last": "Doe", "age": 30}'

    wf_status = dbos.get_workflow_status(wfid)
    assert wf_status is not None
    assert wf_status.status == "SUCCESS"
    assert wf_status.name == "enqueue_test"
    assert wf_status.app_version == GlobalParams.app_version


def test_client_enqueue_wrong_appver(dbos: DBOS, client: DBOSClient) -> None:
    run_client_collateral()

    johnDoe: Person = {"first": "John", "last": "Doe", "age": 30}
    wfid = str(uuid.uuid4())

    options: EnqueueOptions = {
        "queue_name": "test_queue",
        "workflow_name": "enqueue_test",
        "workflow_id": wfid,
        "app_version": "0123456789abcdef",
    }

    client.enqueue(options, 42, "test", johnDoe)

    time.sleep(5)

    wf_status = dbos.get_workflow_status(wfid)
    assert wf_status is not None
    assert wf_status.status == "ENQUEUED"
    assert wf_status.name == "enqueue_test"
    assert wf_status.app_version == options["app_version"]


def test_client_enqueue_idempotent(config: DBOSConfig, client: DBOSClient) -> None:
    DBOS.destroy(destroy_registry=True)

    johnDoe: Person = {"first": "John", "last": "Doe", "age": 30}
    wfid = str(uuid.uuid4())

    options: EnqueueOptions = {
        "queue_name": "test_queue",
        "workflow_name": "enqueue_test",
        "workflow_id": wfid,
    }

    client.enqueue(options, 42, "test", johnDoe)
    client.enqueue(options, 42, "test", johnDoe)

    DBOS(config=config)
    DBOS.launch()
    run_client_collateral()

    handle: WorkflowHandle[str] = DBOS.retrieve_workflow(wfid)
    result = handle.get_result()
    assert result == '42-test-{"first": "John", "last": "Doe", "age": 30}'

    wf_status = DBOS.retrieve_workflow(handle.get_workflow_id()).get_status()
    assert wf_status is not None
    assert wf_status.status == "SUCCESS"
    assert wf_status.name == "enqueue_test"
    assert wf_status.app_version == GlobalParams.app_version

    DBOS.destroy(destroy_registry=True)


def test_client_send_with_topic(client: DBOSClient, dbos: DBOS) -> None:

    from tests.client_collateral import send_test

    run_client_collateral()

    now = time.time_ns()
    wfid = str(uuid.uuid4())
    topic = f"test-topic-{now}"
    message = f"Hello, DBOS! {now}"

    with SetWorkflowID(wfid):
        handle = DBOS.start_workflow(send_test, topic)

    client.send(handle.get_workflow_id(), message, topic)

    result = handle.get_result()
    assert result == message


def test_client_send_no_topic(client: DBOSClient, dbos: DBOS) -> None:

    run_client_collateral()

    now = time.time_ns()
    wfid = str(uuid.uuid4())
    message = f"Hello, DBOS! {now}"

    with SetWorkflowID(wfid):
        handle = DBOS.start_workflow(send_test)

    client.send(handle.get_workflow_id(), message)

    result = handle.get_result()
    assert result == message


def run_send_worker(wfid: str, topic: Optional[str], app_ver: str) -> None:
    script_path = os.path.join(os.path.dirname(__file__), "client_worker.py")
    args = [sys.executable, script_path, wfid]
    if topic is not None:
        args.append(topic)

    env = os.environ.copy()
    env["DBOS__APPVERSION"] = app_ver
    result = subprocess.run(args, env=env, capture_output=True, text=True)
    assert result.returncode == 0, f"Worker failed with error: {result.stderr}"
    DBOS.logger.info(result.stdout)


def test_client_send_idempotent(
    dbos: DBOS, client: DBOSClient, skip_with_sqlite: None
) -> None:
    run_client_collateral()

    now = math.floor(time.time())
    wfid = f"test-send-{now}"
    topic = f"test-topic-{now}"
    message = f"Hello, DBOS! {now}"
    idempotency_key = f"test-idempotency-{now}"
    sendWFID = f"{wfid}-{idempotency_key}"

    run_send_worker(wfid, topic, GlobalParams.app_version)

    client.send(wfid, message, topic, idempotency_key)
    client.send(wfid, message, topic, idempotency_key)

    with dbos._sys_db.engine.connect() as conn:
        nresult = conn.execute(
            sa.text(
                "SELECT * FROM dbos.notifications WHERE destination_uuid = :wfid"
            ).bindparams(wfid=wfid)
        ).fetchall()
        assert len(nresult) == 1
        oresult = conn.execute(
            sa.text(
                "SELECT * FROM dbos.operation_outputs WHERE workflow_uuid = :wfid"
            ).bindparams(wfid=sendWFID)
        ).fetchall()
        assert len(oresult) == 1
        sresult = conn.execute(
            sa.text(
                "SELECT * FROM dbos.workflow_status WHERE workflow_uuid = :wfid"
            ).bindparams(wfid=sendWFID)
        ).fetchall()
        assert len(sresult) == 1

    DBOS._recover_pending_workflows()
    handle: WorkflowHandle[str] = DBOS.retrieve_workflow(wfid)
    result2 = handle.get_result()
    assert result2 == message


def test_client_send_failure(
    dbos: DBOS, client: DBOSClient, skip_with_sqlite: None
) -> None:
    run_client_collateral()

    now = math.floor(time.time())
    wfid = f"test-send-fail-{now}"
    topic = f"test-topic-{now}"
    message = f"Hello, DBOS! {now}"
    idempotency_key = f"test-idempotency-{now}"
    sendWFID = f"{wfid}-{idempotency_key}"

    run_send_worker(wfid, topic, GlobalParams.app_version)

    client.send(wfid, message, topic, idempotency_key)

    # simulate a send failure by deleting notification but leaving the WF status table result
    with dbos._sys_db.engine.connect() as conn:
        oresult = conn.execute(
            sa.text(
                "DELETE FROM dbos.operation_outputs WHERE workflow_uuid = :wfid"
            ).bindparams(wfid=sendWFID)
        )
        assert oresult.rowcount == 1
        nresult = conn.execute(
            sa.text(
                "DELETE FROM dbos.notifications WHERE destination_uuid = :wfid"
            ).bindparams(wfid=wfid)
        )
        assert nresult.rowcount == 1
        sresult = conn.execute(
            sa.text(
                "SELECT recovery_attempts FROM dbos.workflow_status WHERE workflow_uuid = :wfid"
            ).bindparams(wfid=sendWFID)
        ).fetchall()
        assert len(sresult) == 1
        assert sresult[0][0] == 1

    client.send(wfid, message, topic, idempotency_key)

    with dbos._sys_db.engine.connect() as conn:
        s2result = conn.execute(
            sa.text(
                "SELECT recovery_attempts FROM dbos.workflow_status WHERE workflow_uuid = :wfid"
            ).bindparams(wfid=sendWFID)
        ).fetchall()
        assert len(s2result) == 1
        assert s2result[0][0] == 2

    DBOS._recover_pending_workflows()
    handle: WorkflowHandle[str] = DBOS.retrieve_workflow(wfid)
    result = handle.get_result()
    assert result == message


def test_client_get_event(client: DBOSClient, dbos: DBOS) -> None:
    run_client_collateral()

    now = math.floor(time.time())
    wfid = f"test-client-event-{now}"
    key = f"key-{now}"
    value = f"value-{now}"

    with SetWorkflowID(wfid):
        handle = DBOS.start_workflow(event_test, key, value, None)

    client_value = client.get_event(wfid, key, 10)
    assert client_value == value
    result = handle.get_result()
    assert result == f"{key}-{value}"


def test_client_get_event_finished(client: DBOSClient, dbos: DBOS) -> None:
    run_client_collateral()

    now = math.floor(time.time())
    wfid = f"test-client-event-{now}"
    key = f"key-{now}"
    value = f"value-{now}"

    with SetWorkflowID(wfid):
        handle = DBOS.start_workflow(event_test, key, value, None)
        result = handle.get_result()
        assert result == f"{key}-{value}"

    client_value = client.get_event(wfid, key, 10)
    assert client_value == value


def test_client_get_event_update(client: DBOSClient, dbos: DBOS) -> None:
    run_client_collateral()

    now = math.floor(time.time())
    wfid = f"test-client-event-{now}"
    key = f"key-{now}"
    value = f"value-{now}"

    with SetWorkflowID(wfid):
        handle = DBOS.start_workflow(event_test, key, value, 10)

    client_value = client.get_event(wfid, key, 10)
    assert client_value == value
    result = handle.get_result()
    assert result == f"{key}-{value}"
    client_value = client.get_event(wfid, key, 10)
    assert client_value == f"updated-{value}"


def test_client_get_event_update_finished(client: DBOSClient, dbos: DBOS) -> None:
    run_client_collateral()

    now = math.floor(time.time())
    wfid = f"test-client-event-{now}"
    key = f"key-{now}"
    value = f"value-{now}"

    with SetWorkflowID(wfid):
        handle = DBOS.start_workflow(event_test, key, value, 10)
        result = handle.get_result()
        assert result == f"{key}-{value}"

    client_value = client.get_event(wfid, key, 10)
    assert client_value == f"updated-{value}"


def test_client_retrieve_wf(client: DBOSClient, dbos: DBOS) -> None:
    run_client_collateral()

    message = f"Hello, DBOS! {time.time_ns()}"
    handle1 = DBOS.start_workflow(retrieve_test, message)

    handle2: WorkflowHandle[str] = client.retrieve_workflow(handle1.get_workflow_id())
    assert handle1.get_workflow_id() == handle2.get_workflow_id()
    status = handle2.get_status()
    assert status.status == "PENDING"

    result = handle2.get_result()
    assert result == message


def test_client_retrieve_wf_done(client: DBOSClient, dbos: DBOS) -> None:
    run_client_collateral()

    message = f"Hello, DBOS! {time.time_ns()}"
    handle1 = DBOS.start_workflow(retrieve_test, message)
    result1 = handle1.get_result()
    assert result1 == message

    handle2: WorkflowHandle[str] = client.retrieve_workflow(handle1.get_workflow_id())
    assert handle1.get_workflow_id() == handle2.get_workflow_id()
    result2 = handle2.get_result()
    assert result2 == message


def test_client_fork(dbos: DBOS, client: DBOSClient) -> None:
    run_client_collateral()

    options: EnqueueOptions = {
        "queue_name": "test_queue",
        "workflow_name": "fork_test",
    }

    input = 5
    handle: WorkflowHandle[int] = client.enqueue(options, input)
    assert handle.get_result() == input * 2
    assert len(client.list_workflow_steps(handle.workflow_id)) == 2

    fork_id = str(uuid.uuid4())
    with SetWorkflowID(fork_id):
        forked_handle: WorkflowHandle[int] = client.fork_workflow(handle.workflow_id, 1)
    assert forked_handle.workflow_id == fork_id
    assert forked_handle.get_result() == input * 2

    forked_handle = client.fork_workflow(handle.workflow_id, 2)
    assert (
        forked_handle.workflow_id != handle.workflow_id
        and forked_handle.workflow_id != fork_id
    )
    assert forked_handle.get_result() == input * 2

    assert len(client.list_workflows()) == 3


@pytest.mark.asyncio
async def test_client_fork_async(dbos: DBOS, client: DBOSClient) -> None:
    run_client_collateral()

    options: EnqueueOptions = {
        "queue_name": "test_queue",
        "workflow_name": "fork_test",
    }

    input = 5
    handle: WorkflowHandleAsync[int] = await client.enqueue_async(options, input)
    assert await handle.get_result() == input * 2
    assert len(await client.list_workflow_steps_async(handle.workflow_id)) == 2

    forked_handle: WorkflowHandleAsync[int] = await client.fork_workflow_async(
        handle.workflow_id, 1
    )
    assert forked_handle.workflow_id != handle.workflow_id
    assert await forked_handle.get_result() == input * 2

    forked_handle = await client.fork_workflow_async(handle.workflow_id, 2)
    assert forked_handle.workflow_id != handle.workflow_id
    assert await forked_handle.get_result() == input * 2

    assert len(await client.list_workflows_async()) == 3


def test_enqueue_with_deduplication(dbos: DBOS, client: DBOSClient) -> None:
    run_client_collateral()

    wfid = str(uuid.uuid4())
    dedup_id = f"dedup-{wfid}"
    options: EnqueueOptions = {
        "queue_name": "test_queue",
        "workflow_name": "retrieve_test",
        "workflow_id": wfid,
        "deduplication_id": dedup_id,
    }

    handle: WorkflowHandle[str] = client.enqueue(options, "abc")
    # Enqueue itself again should not raise an error
    handle2: WorkflowHandle[str] = client.enqueue(options, "def")

    # Enqueue with the same deduplication ID but different workflow ID should raise an error
    wfid2 = str(uuid.uuid4())
    options["workflow_id"] = wfid2
    with pytest.raises(Exception) as exc_info:
        client.enqueue(options, "def")
    assert (
        f"Workflow {wfid2} was deduplicated due to an existing workflow in queue test_queue with deduplication ID {dedup_id}."
        in str(exc_info.value)
    )

    list_results = client.list_queued_workflows()
    assert len(list_results) == 1
    assert list_results[0].workflow_id == wfid
    assert list_results[0].status in ["PENDING", "ENQUEUED"]
    assert list_results[0].input is not None
    assert list_results[0].output is None

    # Skip loading input
    list_results = client.list_queued_workflows(load_input=False)
    assert len(list_results) == 1
    assert list_results[0].workflow_id == wfid
    assert list_results[0].status in ["PENDING", "ENQUEUED"]
    assert list_results[0].input is None
    assert list_results[0].output is None

    assert handle.get_result() == "abc"
    assert handle2.get_result() == "abc"


def test_enqueue_with_priority(dbos: DBOS, client: DBOSClient) -> None:
    importlib.reload(client_collateral)
    from tests.client_collateral import inorder_results

    options: EnqueueOptions = {
        "queue_name": "inorder_queue",
        "workflow_name": "retrieve_test",
        "priority": -1,
    }

    with pytest.raises(Exception) as exc_info:
        client.enqueue(options, "abc")
    assert "Invalid priority" in str(exc_info.value)

    # Without priority should work
    del options["priority"]
    handle: WorkflowHandle[str] = client.enqueue(options, "abc")

    # Enqueue with a lower priority
    options["priority"] = 5
    handle2: WorkflowHandle[str] = client.enqueue(options, "def")

    # Enqueue with a higher priority
    options["priority"] = 1
    handle3: WorkflowHandle[str] = client.enqueue(options, "ghi")

    assert handle.get_result() == "abc"
    assert handle3.get_result() == "ghi"
    assert handle2.get_result() == "def"

    # Should be in the order of priority
    assert inorder_results == ["abc", "ghi", "def"]


def test_client_bad_url() -> None:
    with pytest.raises(DBAPIError) as exc_info:
        DBOSClient("postgresql://postgres:fakepassword@localhost:5433/fake_database")



"""dbos_migrations

Revision ID: 471b60d64126
Revises: 01ce9f07bd10
Create Date: 2025-08-21 14:22:31.455266

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "471b60d64126"
down_revision: Union[str, None] = "01ce9f07bd10"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # Create dbos_migrations table
    op.create_table(
        "dbos_migrations",
        sa.Column("version", sa.BigInteger(), nullable=False),
        sa.PrimaryKeyConstraint("version"),
        schema="dbos",
    )

    # Insert initial version 1
    op.execute("INSERT INTO dbos.dbos_migrations (version) VALUES (1)")


def downgrade() -> None:
    op.drop_table("dbos_migrations", schema="dbos")



import threading
import time
import uuid
from concurrent.futures import Future, ThreadPoolExecutor
from typing import Tuple, cast

from sqlalchemy import text

# Public API
from dbos import DBOS, SetWorkflowID
from tests.conftest import using_sqlite


def test_concurrent_workflows(dbos: DBOS) -> None:
    @DBOS.workflow()
    def test_workflow() -> str:
        time.sleep(1)
        workflow_id = DBOS.workflow_id
        assert workflow_id is not None
        return workflow_id

    def test_thread(id: str) -> str:
        with SetWorkflowID(id):
            return test_workflow()

    num_threads = 10
    with ThreadPoolExecutor(max_workers=num_threads) as executor:
        futures: list[Tuple[str, Future[str]]] = []
        for _ in range(num_threads):
            id = str(uuid.uuid4())
            futures.append((id, executor.submit(test_thread, id)))
        for id, future in futures:
            assert id == future.result()


def test_concurrent_conflict_uuid(dbos: DBOS) -> None:
    condition = threading.Condition()
    step_count = 0
    txn_count = 0

    @DBOS.step()
    def test_step() -> str:
        nonlocal step_count
        condition.acquire()
        step_count += 1
        if step_count % 2 == 1:
            # Wait for the other one to notify
            condition.wait()
        else:
            # Notify the other one
            condition.notify()
        condition.release()

        workflow_id = DBOS.workflow_id
        assert workflow_id is not None
        return workflow_id

    @DBOS.workflow()
    def test_workflow() -> str:
        res = test_step()
        return res

    # Need to set isolation level to a lower one, otherwise it gets serialization error instead (we already handle it correctly by automatic retries).
    @DBOS.transaction(
        isolation_level=("SERIALIZABLE" if using_sqlite() else "REPEATABLE READ")
    )
    def test_transaction() -> str:
        DBOS.sql_session.execute(text("SELECT 1")).fetchall()
        nonlocal txn_count
        condition.acquire()
        txn_count += 1
        if txn_count % 2 == 1:
            # Wait for the other one to notify
            condition.wait()
        else:
            # Notify the other one
            condition.notify()
        condition.release()

        workflow_id = DBOS.workflow_id
        assert workflow_id is not None
        return workflow_id

    def test_txn_thread(id: str) -> str:
        with SetWorkflowID(id):
            return test_transaction()

    wfuuid = str(uuid.uuid4())
    with SetWorkflowID(wfuuid):
        wf_handle1 = dbos.start_workflow(test_workflow)

    with SetWorkflowID(wfuuid):
        wf_handle2 = dbos.start_workflow(test_workflow)

    # These two workflows should run concurrently, but both should succeed.
    assert wf_handle1.get_result() == wfuuid
    assert wf_handle2.get_result() == wfuuid

    # Make sure temp transactions can handle conflicts as well.
    wfuuid = str(uuid.uuid4())
    with ThreadPoolExecutor(max_workers=2) as executor:
        future1 = executor.submit(test_txn_thread, wfuuid)
        future2 = executor.submit(test_txn_thread, wfuuid)

    assert future1.result() == wfuuid
    assert future2.result() == wfuuid


def test_concurrent_recv(dbos: DBOS) -> None:
    condition = threading.Condition()
    counter = 0

    @DBOS.workflow()
    def test_workflow(topic: str) -> str:
        nonlocal counter
        condition.acquire()
        counter += 1
        if counter % 2 == 1:
            # Wait for the other one to notify
            condition.wait()
        else:
            # Notify the other one
            condition.notify()
        condition.release()
        m = cast(str, DBOS.recv(topic, 5))
        return m

    def test_thread(id: str, topic: str) -> str:
        with SetWorkflowID(id):
            return test_workflow(topic)

    wfuuid = str(uuid.uuid4())
    topic = "test_topic"
    with ThreadPoolExecutor(max_workers=2) as executor:
        future1 = executor.submit(test_thread, wfuuid, topic)
        future2 = executor.submit(test_thread, wfuuid, topic)

        expected_message = "test message"
        DBOS.send(wfuuid, expected_message, topic)
        # Both should return the same message
        assert future1.result() == future2.result()
        assert future1.result() == expected_message
        # Make sure the notification map is empty
        assert not dbos._sys_db.notifications_map._dict


def test_concurrent_getevent(dbos: DBOS) -> None:
    @DBOS.workflow()
    def test_workflow(event_name: str, value: str) -> str:
        DBOS.set_event(event_name, value)
        return value

    def test_thread(id: str, event_name: str) -> str:
        return cast(str, DBOS.get_event(id, event_name, 5))

    wfuuid = str(uuid.uuid4())
    event_name = "test_event"
    with ThreadPoolExecutor(max_workers=2) as executor:
        future1 = executor.submit(test_thread, wfuuid, event_name)
        future2 = executor.submit(test_thread, wfuuid, event_name)

        expected_message = "test message"
        with SetWorkflowID(wfuuid):
            test_workflow(event_name, expected_message)

        # Both should return the same message
        assert future1.result() == future2.result()
        assert future1.result() == expected_message
        # Make sure the event map is empty
        assert not dbos._sys_db.workflow_events_map._dict



from dataclasses import dataclass, field
from typing import Optional, Tuple

import pytest
from fastapi import FastAPI
from fastapi.testclient import TestClient
from inline_snapshot import snapshot
from opentelemetry._logs import set_logger_provider
from opentelemetry.sdk import trace as tracesdk
from opentelemetry.sdk._logs import LoggerProvider, LoggingHandler
from opentelemetry.sdk._logs.export import BatchLogRecordProcessor, InMemoryLogExporter
from opentelemetry.sdk.trace.export import SimpleSpanProcessor
from opentelemetry.sdk.trace.export.in_memory_span_exporter import InMemorySpanExporter
from opentelemetry.trace.span import format_trace_id

from dbos import DBOS, DBOSConfig
from dbos._logger import dbos_logger
from dbos._tracer import dbos_tracer
from dbos._utils import GlobalParams


@dataclass
class BasicSpan:
    content: str
    children: list["BasicSpan"] = field(default_factory=list)
    parent_id: Optional[int] = field(repr=False, compare=False, default=None)


def test_spans(config: DBOSConfig) -> None:
    exporter = InMemorySpanExporter()
    span_processor = SimpleSpanProcessor(exporter)
    provider = tracesdk.TracerProvider()
    provider.add_span_processor(span_processor)
    dbos_tracer.set_provider(provider)

    DBOS.destroy(destroy_registry=True)
    config["otlp_attributes"] = {"foo": "bar"}
    DBOS(config=config)
    DBOS.launch()

    my_tracer = provider.get_tracer("dbos")

    @DBOS.workflow()
    def test_workflow() -> None:
        with my_tracer.start_as_current_span(  # pyright: ignore[reportAttributeAccessIssue]
            "manual_span"
        ):
            test_step()
            current_span = DBOS.span
            subspan = DBOS.tracer.start_span(
                {"name": "a new span"}, parent=current_span
            )
            # Note: DBOS.tracer.start_span() does not set the new span as the current span. So this log is still attached to the workflow span.
            DBOS.logger.info("This is a test_workflow")
            subspan.add_event("greeting_event", {"name": "a new event"})
            DBOS.tracer.end_span(subspan)

    @DBOS.step()
    def test_step() -> None:
        DBOS.logger.info("This is a test_step")
        return

    # Set up in-memory log exporter
    log_exporter = InMemoryLogExporter()  # type: ignore
    log_processor = BatchLogRecordProcessor(log_exporter)
    log_provider = LoggerProvider()
    log_provider.add_log_record_processor(log_processor)
    set_logger_provider(log_provider)
    dbos_logger.addHandler(LoggingHandler(logger_provider=log_provider))

    test_workflow()

    log_processor.force_flush(timeout_millis=5000)
    logs = log_exporter.get_finished_logs()
    assert len(logs) == 2
    for log in logs:
        assert log.log_record.attributes is not None
        assert (
            log.log_record.attributes["applicationVersion"] == GlobalParams.app_version
        )
        assert log.log_record.attributes["executorID"] == GlobalParams.executor_id
        assert log.log_record.attributes["foo"] == "bar"
        # Make sure the log record has a span_id and trace_id
        assert log.log_record.span_id is not None and log.log_record.span_id > 0
        assert log.log_record.trace_id is not None and log.log_record.trace_id > 0
        assert (
            log.log_record.body == "This is a test_step"
            or log.log_record.body == "This is a test_workflow"
        )
        assert log.log_record.attributes["traceId"] == format_trace_id(
            log.log_record.trace_id
        )

    spans = exporter.get_finished_spans()

    for span in spans:
        if span.name == "manual_span":
            # Skip the manual span because it was not created by DBOS.tracer
            continue
        assert span.attributes is not None
        assert span.attributes["applicationVersion"] == GlobalParams.app_version
        assert span.attributes["executorID"] == GlobalParams.executor_id
        assert span.context is not None
        assert span.attributes["foo"] == "bar"
        assert span.context.span_id > 0
        assert span.context.trace_id > 0

    assert spans[0].name == test_step.__qualname__
    assert spans[1].name == "a new span"
    assert spans[3].name == test_workflow.__qualname__

    assert spans[0].parent.span_id == spans[2].context.span_id  # type: ignore
    assert spans[1].parent.span_id == spans[2].context.span_id  # type: ignore
    assert spans[2].parent.span_id == spans[3].context.span_id  # type: ignore
    assert spans[3].parent == None

    # Span ID and trace ID should match the log record
    # For pyright
    assert spans[0].context is not None
    assert spans[2].context is not None
    assert logs[0].log_record.span_id == spans[0].context.span_id
    assert logs[0].log_record.trace_id == spans[0].context.trace_id
    assert logs[1].log_record.span_id == spans[2].context.span_id
    assert logs[1].log_record.trace_id == spans[2].context.trace_id

    # Test the span tree structure
    basic_spans = {
        span.context.span_id: BasicSpan(  # pyright: ignore[reportOptionalMemberAccess]
            content=span.name, parent_id=span.parent.span_id if span.parent else None
        )
        for span in spans
    }
    root_span = None
    for basic_span in basic_spans.values():
        if basic_span.parent_id is None:
            root_span = basic_span
        else:
            parent_id = basic_span.parent_id
            parent_span = basic_spans[parent_id]
            parent_span.children.append(basic_span)

    assert len(spans) == 4
    # Make sure the span tree structure is correct
    assert root_span == snapshot(
        BasicSpan(
            content="test_spans.<locals>.test_workflow",
            children=[
                BasicSpan(
                    content="manual_span",
                    children=[
                        BasicSpan(content="test_spans.<locals>.test_step"),
                        BasicSpan(content="a new span"),
                    ],
                )
            ],
        )
    )


@pytest.mark.asyncio
async def test_spans_async(dbos: DBOS) -> None:
    exporter = InMemorySpanExporter()
    span_processor = SimpleSpanProcessor(exporter)
    provider = tracesdk.TracerProvider()
    provider.add_span_processor(span_processor)
    dbos_tracer.set_provider(provider)

    my_tracer = provider.get_tracer("dbos")

    @DBOS.workflow()
    async def test_workflow() -> None:
        with my_tracer.start_as_current_span(  # pyright: ignore[reportAttributeAccessIssue]
            "manual_span"
        ):
            await test_step()
            current_span = DBOS.span
            subspan = DBOS.tracer.start_span(
                {"name": "a new span"}, parent=current_span
            )
            # Note: DBOS.tracer.start_span() does not set the new span as the current span. So this log is still attached to the workflow span.
            DBOS.logger.info("This is a test_workflow")
            subspan.add_event("greeting_event", {"name": "a new event"})
            DBOS.tracer.end_span(subspan)

    @DBOS.step()
    async def test_step() -> None:
        DBOS.logger.info("This is a test_step")
        return

    # Set up in-memory log exporter
    log_exporter = InMemoryLogExporter()  # type: ignore
    log_processor = BatchLogRecordProcessor(log_exporter)
    log_provider = LoggerProvider()
    log_provider.add_log_record_processor(log_processor)
    set_logger_provider(log_provider)
    dbos_logger.addHandler(LoggingHandler(logger_provider=log_provider))

    await test_workflow()

    log_processor.force_flush(timeout_millis=5000)
    logs = log_exporter.get_finished_logs()
    assert len(logs) == 2
    for log in logs:
        assert log.log_record.attributes is not None
        assert (
            log.log_record.attributes["applicationVersion"] == GlobalParams.app_version
        )
        assert log.log_record.attributes["executorID"] == GlobalParams.executor_id
        # Make sure the log record has a span_id and trace_id
        assert log.log_record.span_id is not None and log.log_record.span_id > 0
        assert log.log_record.trace_id is not None and log.log_record.trace_id > 0
        assert (
            log.log_record.body == "This is a test_step"
            or log.log_record.body == "This is a test_workflow"
        )
        assert log.log_record.attributes["traceId"] == format_trace_id(
            log.log_record.trace_id
        )

    spans = exporter.get_finished_spans()

    assert len(spans) == 4

    for span in spans:
        if span.name == "manual_span":
            # Skip the manual span because it was not created by DBOS.tracer
            continue
        assert span.attributes is not None
        assert span.attributes["applicationVersion"] == GlobalParams.app_version
        assert span.attributes["executorID"] == GlobalParams.executor_id
        assert span.context is not None
        assert span.context.span_id > 0
        assert span.context.trace_id > 0

    assert spans[0].name == test_step.__qualname__
    assert spans[1].name == "a new span"
    assert spans[3].name == test_workflow.__qualname__

    assert spans[0].parent.span_id == spans[2].context.span_id  # type: ignore
    assert spans[1].parent.span_id == spans[2].context.span_id  # type: ignore
    assert spans[2].parent.span_id == spans[3].context.span_id  # type: ignore
    assert spans[3].parent == None

    # Span ID and trace ID should match the log record
    assert spans[0].context is not None
    assert spans[2].context is not None
    assert logs[0].log_record.span_id == spans[0].context.span_id
    assert logs[0].log_record.trace_id == spans[0].context.trace_id
    assert logs[1].log_record.span_id == spans[2].context.span_id
    assert logs[1].log_record.trace_id == spans[2].context.trace_id

    # Test the span tree structure
    basic_spans = {
        span.context.span_id: BasicSpan(  # pyright: ignore[reportOptionalMemberAccess]
            content=span.name, parent_id=span.parent.span_id if span.parent else None
        )
        for span in spans
    }
    root_span = None
    for basic_span in basic_spans.values():
        if basic_span.parent_id is None:
            root_span = basic_span
        else:
            parent_id = basic_span.parent_id
            parent_span = basic_spans[parent_id]
            parent_span.children.append(basic_span)

    assert len(spans) == 4
    # Make sure the span tree structure is correct
    assert root_span == snapshot(
        BasicSpan(
            content="test_spans_async.<locals>.test_workflow",
            children=[
                BasicSpan(
                    content="manual_span",
                    children=[
                        BasicSpan(content="test_spans_async.<locals>.test_step"),
                        BasicSpan(content="a new span"),
                    ],
                )
            ],
        )
    )


def test_wf_fastapi(dbos_fastapi: Tuple[DBOS, FastAPI]) -> None:
    dbos, app = dbos_fastapi

    @app.get("/wf")
    @DBOS.workflow()
    def test_workflow_endpoint() -> str:
        dbos.logger.info("This is a test_workflow_endpoint")
        return "test"

    exporter = InMemorySpanExporter()
    span_processor = SimpleSpanProcessor(exporter)
    provider = tracesdk.TracerProvider()
    provider.add_span_processor(span_processor)
    dbos_tracer.set_provider(provider)

    # Set up in-memory log exporter
    log_exporter = InMemoryLogExporter()  # type: ignore
    log_processor = BatchLogRecordProcessor(log_exporter)
    log_provider = LoggerProvider()
    log_provider.add_log_record_processor(log_processor)
    set_logger_provider(log_provider)
    dbos_logger.addHandler(LoggingHandler(logger_provider=log_provider))

    client = TestClient(app)
    response = client.get("/wf")
    assert response.status_code == 200
    assert response.text == '"test"'

    log_processor.force_flush(timeout_millis=5000)
    logs = log_exporter.get_finished_logs()
    assert len(logs) == 1
    assert logs[0].log_record.attributes is not None
    assert (
        logs[0].log_record.attributes["applicationVersion"] == GlobalParams.app_version
    )
    assert logs[0].log_record.span_id is not None and logs[0].log_record.span_id > 0
    assert logs[0].log_record.trace_id is not None and logs[0].log_record.trace_id > 0
    assert logs[0].log_record.body == "This is a test_workflow_endpoint"
    assert logs[0].log_record.attributes["traceId"] == format_trace_id(
        logs[0].log_record.trace_id
    )

    spans = exporter.get_finished_spans()

    assert len(spans) == 2

    for span in spans:
        assert span.attributes is not None
        assert span.attributes["applicationVersion"] == GlobalParams.app_version
        assert span.context is not None
        assert span.context.span_id > 0
        assert span.context.trace_id > 0

    assert spans[0].name == test_workflow_endpoint.__qualname__
    assert spans[1].name == "/wf"
    assert spans[1].attributes is not None
    assert spans[1].attributes["responseCode"] == 200

    assert spans[0].parent.span_id == spans[1].context.span_id  # type: ignore
    assert spans[1].parent == None

    # Span ID and trace ID should match the log record
    assert spans[0].context is not None
    assert logs[0].log_record.span_id == spans[0].context.span_id
    assert logs[0].log_record.trace_id == spans[0].context.trace_id


def test_disable_otlp_no_spans(config: DBOSConfig) -> None:
    DBOS.destroy(destroy_registry=True)
    config["otlp_attributes"] = {"foo": "bar"}
    config["disable_otlp"] = True
    DBOS(config=config)
    DBOS.launch()

    @DBOS.workflow()
    def test_workflow() -> None:
        test_step()
        DBOS.logger.info("This is a test_workflow")

    @DBOS.step()
    def test_step() -> None:
        DBOS.logger.info("This is a test_step")
        return

    exporter = InMemorySpanExporter()
    span_processor = SimpleSpanProcessor(exporter)
    provider = tracesdk.TracerProvider()
    provider.add_span_processor(span_processor)
    dbos_tracer.set_provider(provider)

    # Set up in-memory log exporter
    log_exporter = InMemoryLogExporter()  # type: ignore
    log_processor = BatchLogRecordProcessor(log_exporter)
    log_provider = LoggerProvider()
    log_provider.add_log_record_processor(log_processor)
    set_logger_provider(log_provider)
    dbos_logger.addHandler(LoggingHandler(logger_provider=log_provider))

    test_workflow()

    log_processor.force_flush(timeout_millis=5000)
    logs = log_exporter.get_finished_logs()
    assert len(logs) == 2
    for log in logs:
        assert log.log_record.attributes is not None
        assert (
            log.log_record.attributes["applicationVersion"] == GlobalParams.app_version
        )
        assert log.log_record.attributes["executorID"] == GlobalParams.executor_id
        assert log.log_record.attributes["foo"] == "bar"
        # We disable OTLP, so no span_id or trace_id should be present
        assert log.log_record.span_id is not None and log.log_record.span_id == 0
        assert log.log_record.trace_id is not None and log.log_record.trace_id == 0
        assert (
            log.log_record.body == "This is a test_step"
            or log.log_record.body == "This is a test_workflow"
        )
        assert log.log_record.attributes.get("traceId") is None

    spans = exporter.get_finished_spans()

    # No spans should be created since OTLP is disabled
    assert len(spans) == 0



import time

import pytest

# Public API
from dbos import DBOS, SetWorkflowID, WorkflowHandle

# Private API used because this is a test
from dbos._context import DBOSContextEnsure, assert_current_dbos_context
from tests.conftest import default_config


def test_dbos_singleton(cleanup_test_databases: None) -> None:
    # Initialize singleton
    DBOS.destroy()  # In case of other tests leaving it

    # Simulate an app that does some imports of its own code, then defines DBOS,
    #    then imports more
    from tests.classdefs import (
        DBOSSendRecv,
        DBOSTestClass,
        DBOSTestRoles,
        DBOSTestWrapperMethods,
    )

    dbos: DBOS = DBOS(config=default_config())

    from tests.more_classdefs import DBOSWFEvents, wfFunc

    DBOS.launch()  # Usually framework (fastapi) does this via lifecycle event

    # Basics
    with SetWorkflowID("wfid"):
        res = wfFunc("f")
    assert res == "f1"

    res = DBOSTestClass.test_workflow_cls("a", "b")
    assert res == "b1a"

    inst = DBOSTestClass()
    res = inst.test_workflow("c", "d")
    assert res == "d1c"

    wh = DBOS.start_workflow(DBOSTestClass.test_workflow_cls, "a", "b")
    assert wh.get_result() == "b1a"

    wh = DBOS.start_workflow(inst.test_workflow, "c", "d")
    assert wh.get_result() == "d1c"
    stati = DBOS.get_workflow_status(wh.get_workflow_id())
    assert stati
    assert stati.config_name == "myconfig"
    assert stati.class_name and "DBOSTestClass" in stati.class_name
    wfhr: WorkflowHandle[str] = DBOS.retrieve_workflow(wh.get_workflow_id())
    assert wfhr.workflow_id == wh.get_workflow_id()

    # Test wrapper class methods
    wrapper_inst = DBOSTestWrapperMethods("dynamicconfig")
    res = wrapper_inst.test_workflow("x", "y")
    assert res == "y1x"

    wh = DBOS.start_workflow(wrapper_inst.test_workflow, "x", "y")
    assert wh.get_result() == "y1x"
    stati = DBOS.get_workflow_status(wh.get_workflow_id())
    assert stati
    assert stati.config_name is None
    assert stati.class_name is None
    assert stati.name == "dynamicconfig_test_workflow"
    wfhr = DBOS.retrieve_workflow(wh.get_workflow_id())
    assert wfhr.workflow_id == wh.get_workflow_id()

    stepi = DBOS.list_workflow_steps(wh.get_workflow_id())
    assert len(stepi) == 2
    assert stepi[0]["function_name"] == "dynamicconfig_test_transaction"
    assert stepi[1]["function_name"] == "dynamicconfig_test_step"

    # Roles

    with DBOSContextEnsure():
        with pytest.raises(Exception) as exc_info:
            DBOSTestRoles.greetfunc("Nobody")
        assert "DBOS Error 8" in str(exc_info.value)

        ctx = assert_current_dbos_context()
        ctx.authenticated_roles = ["user", "admin"]
        res = inst.test_func_admin("admin")
        assert res == "myconfig:admin"

        assert DBOSTestRoles.greetfunc("user") == "Hello user"

    # Send / Recv
    dest_uuid = str("sruuid1")

    with SetWorkflowID(dest_uuid):
        handle = dbos.start_workflow(DBOSSendRecv.test_recv_workflow, "testtopic")
        assert handle.get_workflow_id() == dest_uuid

    send_uuid = str("sruuid2")
    with SetWorkflowID(send_uuid):
        res = DBOSSendRecv.test_send_workflow(handle.get_workflow_id(), "testtopic")
        assert res == dest_uuid

    begin_time = time.time()
    assert handle.get_result() == "test2-test1-test3"
    duration = time.time() - begin_time
    assert duration < 3.0

    # Events
    wfuuid = str("sendwf1")
    with SetWorkflowID(wfuuid):
        DBOSWFEvents.test_setevent_workflow()
    with SetWorkflowID(wfuuid):
        DBOSWFEvents.test_setevent_workflow()

    value1 = DBOSWFEvents.test_getevent_workflow(wfuuid, "key1")
    assert value1 == "value1"

    value2 = DBOSWFEvents.test_getevent_workflow(wfuuid, "key2")
    assert value2 == "value2"

    # Run getEvent outside of a workflow
    value1 = DBOS.get_event(wfuuid, "key1")
    assert value1 == "value1"

    value2 = DBOS.get_event(wfuuid, "key2")
    assert value2 == "value2"

    begin_time = time.time()
    value3 = DBOSWFEvents.test_getevent_workflow(wfuuid, "key3")
    assert value3 is None
    duration = time.time() - begin_time
    assert duration < 1  # None is from the event not from the timeout

    DBOS.destroy()


def test_dbos_singleton_negative(cleanup_test_databases: None) -> None:
    # Initialize singleton
    DBOS.destroy()  # In case of other tests leaving it

    # Simulate an app that does some imports of its own code, then defines DBOS,
    #    then imports more
    from tests.classdefs import DBOSTestClass

    DBOS(config=default_config())

    # Something should have launched
    with pytest.raises(Exception) as exc_info:
        DBOSTestClass.test_workflow_cls("a", "b")
    assert "launch" in str(exc_info.value)

    DBOS.destroy()



import time
import uuid
from datetime import datetime
from typing import TYPE_CHECKING, List, Optional, Union

from dbos._context import get_local_dbos_context

from ._app_db import ApplicationDatabase
from ._sys_db import (
    GetQueuedWorkflowsInput,
    GetWorkflowsInput,
    StepInfo,
    SystemDatabase,
    WorkflowStatus,
    WorkflowStatusString,
)

if TYPE_CHECKING:
    from ._dbos import DBOS


def list_workflows(
    sys_db: SystemDatabase,
    *,
    workflow_ids: Optional[List[str]] = None,
    status: Optional[Union[str, List[str]]] = None,
    start_time: Optional[str] = None,
    end_time: Optional[str] = None,
    name: Optional[str] = None,
    app_version: Optional[str] = None,
    user: Optional[str] = None,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
    sort_desc: bool = False,
    workflow_id_prefix: Optional[str] = None,
    load_input: bool = True,
    load_output: bool = True,
) -> List[WorkflowStatus]:
    input = GetWorkflowsInput()
    input.workflow_ids = workflow_ids
    input.authenticated_user = user
    input.start_time = start_time
    input.end_time = end_time
    input.status = status if status is None or isinstance(status, list) else [status]
    input.application_version = app_version
    input.limit = limit
    input.name = name
    input.offset = offset
    input.sort_desc = sort_desc
    input.workflow_id_prefix = workflow_id_prefix

    infos: List[WorkflowStatus] = sys_db.get_workflows(
        input, load_input=load_input, load_output=load_output
    )

    return infos


def list_queued_workflows(
    sys_db: SystemDatabase,
    *,
    queue_name: Optional[str] = None,
    status: Optional[Union[str, List[str]]] = None,
    start_time: Optional[str] = None,
    end_time: Optional[str] = None,
    name: Optional[str] = None,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
    sort_desc: bool = False,
    load_input: bool = True,
) -> List[WorkflowStatus]:
    input: GetQueuedWorkflowsInput = {
        "queue_name": queue_name,
        "start_time": start_time,
        "end_time": end_time,
        "status": status if status is None or isinstance(status, list) else [status],
        "limit": limit,
        "name": name,
        "offset": offset,
        "sort_desc": sort_desc,
    }

    infos: List[WorkflowStatus] = sys_db.get_queued_workflows(
        input, load_input=load_input
    )
    return infos


def get_workflow(sys_db: SystemDatabase, workflow_id: str) -> Optional[WorkflowStatus]:
    input = GetWorkflowsInput()
    input.workflow_ids = [workflow_id]

    infos: List[WorkflowStatus] = sys_db.get_workflows(input)
    if not infos:
        return None

    return infos[0]


def list_workflow_steps(
    sys_db: SystemDatabase, app_db: ApplicationDatabase, workflow_id: str
) -> List[StepInfo]:
    steps = sys_db.get_workflow_steps(workflow_id)
    transactions = app_db.get_transactions(workflow_id)
    merged_steps = steps + transactions
    merged_steps.sort(key=lambda step: step["function_id"])
    return merged_steps


def fork_workflow(
    sys_db: SystemDatabase,
    app_db: ApplicationDatabase,
    workflow_id: str,
    start_step: int,
    *,
    application_version: Optional[str],
) -> str:

    ctx = get_local_dbos_context()
    if ctx is not None and len(ctx.id_assigned_for_next_workflow) > 0:
        forked_workflow_id = ctx.id_assigned_for_next_workflow
        ctx.id_assigned_for_next_workflow = ""
    else:
        forked_workflow_id = str(uuid.uuid4())
    app_db.clone_workflow_transactions(workflow_id, forked_workflow_id, start_step)
    sys_db.fork_workflow(
        workflow_id,
        forked_workflow_id,
        start_step,
        application_version=application_version,
    )
    return forked_workflow_id


def garbage_collect(
    dbos: "DBOS",
    cutoff_epoch_timestamp_ms: Optional[int],
    rows_threshold: Optional[int],
) -> None:
    if cutoff_epoch_timestamp_ms is None and rows_threshold is None:
        return
    result = dbos._sys_db.garbage_collect(
        cutoff_epoch_timestamp_ms=cutoff_epoch_timestamp_ms,
        rows_threshold=rows_threshold,
    )
    if result is not None:
        cutoff_epoch_timestamp_ms, pending_workflow_ids = result
        dbos._app_db.garbage_collect(cutoff_epoch_timestamp_ms, pending_workflow_ids)


def global_timeout(dbos: "DBOS", cutoff_epoch_timestamp_ms: int) -> None:
    cutoff_iso = datetime.fromtimestamp(cutoff_epoch_timestamp_ms / 1000).isoformat()
    for workflow in dbos.list_workflows(
        status=WorkflowStatusString.PENDING.value, end_time=cutoff_iso
    ):
        dbos.cancel_workflow(workflow.workflow_id)
    for workflow in dbos.list_workflows(
        status=WorkflowStatusString.ENQUEUED.value, end_time=cutoff_iso
    ):
        dbos.cancel_workflow(workflow.workflow_id)



import json
from typing import List, Optional, TypedDict, cast

from dbos import DBOS, Queue


class Person(TypedDict):
    first: str
    last: str
    age: int


queue = Queue("test_queue")
inorder_queue = Queue("inorder_queue", 1, priority_enabled=True)
inorder_results: List[str] = []


@DBOS.workflow()
def enqueue_test(numVal: int, strVal: str, person: Person) -> str:
    return f"{numVal}-{strVal}-{json.dumps(person)}"


@DBOS.workflow()
def send_test(topic: Optional[str] = None) -> str:
    return cast(str, DBOS.recv(topic, 60))


@DBOS.workflow()
def retrieve_test(value: str) -> str:
    DBOS.sleep(5)
    inorder_results.append(value)
    return value


@DBOS.workflow()
def event_test(key: str, value: str, update: Optional[int] = None) -> str:
    DBOS.set_event(key, value)
    if update is not None:
        DBOS.sleep(update)
        DBOS.set_event(key, f"updated-{value}")
    return f"{key}-{value}"


@DBOS.workflow()
def blocked_workflow() -> None:
    while True:
        DBOS.sleep(0.1)


@DBOS.transaction()
def test_txn(x: int) -> int:
    return x


@DBOS.step()
def test_step(x: int) -> int:
    return x


@DBOS.workflow()
def fork_test(x: int) -> int:
    return test_txn(x) + test_step(x)
